from __future__ import annotations

import datetime
import logging
from typing import Optional, TYPE_CHECKING, List, Dict

from vectice.api import Client
from vectice.api.json import RunInput, JobType, RunStatus
from .integration import AbstractIntegration
from .run import Run
from ..api.json.property import create_properties_input

if TYPE_CHECKING:
    from vectice.models import Project
    from vectice import Reference

import inspect


class Job:
    def __init__(
        self, name: str, id: int, project: Project, description: Optional[str] = None, type: JobType = JobType.OTHER
    ):
        """
        :param id:Job identifier
        :param project: The project this job belong to
        :param name: Name of the job
        :param description: Quick description of the job
        :param type: type of Job it can be "EXTRACTION"|"PREPARATION"|"TRAINING"|"INFERENCE"|"DEPLOYMENT"|"OTHER"
        """
        self._id = id
        self._project: Project = project
        self._name: str = name
        self._description: Optional[str] = description
        self._type: JobType = type
        self._client: Client = project._client
        self._integration_client: Optional[AbstractIntegration] = project._integration_client
        self._logger = logging.getLogger(self.__class__.__name__)

    def __repr__(self):
        return f"Job(name={self.name}, id={self.id}, description={self.description}, type={self.type})"

    @property
    def id(self) -> int:
        """
        Job identifier.
        :return: int
        """
        return self._id

    @id.setter
    def id(self, id):
        self._id = id

    @property
    def project(self) -> Project:
        """
        The project this job belongs to.
        :return: Project
        """
        return self._project

    @project.setter
    def project(self, project):
        self._project = project

    @property
    def name(self) -> Optional[str]:
        """
        Name of the job.
        :return: Optional[str]
        """
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def type(self) -> JobType:
        """
        The job type.
        :return: JobType
        """
        return self._type

    @type.setter
    def type(self, job_type):
        self._type = job_type

    @property
    def description(self) -> Optional[str]:
        """
        Quick description of the job.
        :return: Optional[str]
        """
        return self._description

    @description.setter
    def description(self, description):
        self._description = description

    @classmethod
    def from_dict(cls, env):
        return cls(**{k: v for k, v in env.items() if k in inspect.signature(cls).parameters})

    def create_run(
        self,
        name: Optional[str] = None,
        properties: Optional[Dict] = None,
        auto_code: bool = False,
        check_remote_repository: bool = True,
        notes: Optional[str] = None,
        system_name: Optional[str] = None,
    ) -> Run:
        """
        Creates a run in the specified project with the information passed to this method.
        The run is not viewable in the Vectice platform until the job is started.
        See :class:`Run` for more information
        :param name: The run name.
        :param properties: Properties of the run
        :param auto_code: indicate if auto code should be used or not
        :param check_remote_repository: indicate if we should check remote repository on code version creation
        :return: Run
        """
        data = RunInput(
            name=name,
            properties=properties,
            description=notes,
            systemName=system_name,
            startDate=datetime.datetime.utcnow(),
        )
        run_output = self._client.create_run(data, self._id)
        if properties:
            self._client.create_run_properties(run_output.id, create_properties_input(properties))
        updated_properties = self._client.list_run_properties(run_output.id)
        self._logger.info(f"Run with id: {run_output.id} successfully created.")
        return Run(
            run_output.id,
            self,
            run_output.name,
            run_output.system_name,
            run_output.created_date,
            None,
            None,
            run_output.status,
            auto_code=auto_code,
            check_remote_repository=check_remote_repository,
            description=run_output.description,
            properties={prop.key: prop.value for prop in updated_properties.list},
        )

    def delete_run(self, run: Reference) -> None:
        """
        Deletes the specified run.
        The run must be part of this job.

        :param run: The run name or id
        :return: None
        """
        run_id = self._client.get_run(run, self.id).id
        self._client.delete_run(run, self.id)
        self._logger.info(f"Run with id: {run_id} successfully deleted.")

    def get_run(
        self,
        run: Reference,
    ) -> Run:
        """
        Gets a Run.

        :param run: The run name or id
        :return: A Run
        """
        run_output = self._client.get_run(run, self._id)
        properties = self._client.list_run_properties(run_output.id)
        self._logger.info(f"Run with id: {run_output.id} successfully retrieved.")
        return Run(
            run_output.id,
            self,
            run_output.name,
            run_output.system_name,
            run_output.created_date,
            run_output.start_date,
            run_output.end_date,
            run_output.status,
            properties={prop.key: prop.value for prop in properties.list},
        )

    def update_run(
        self,
        run: Reference,
        name: Optional[str] = None,
        system_name: Optional[str] = None,
        status: Optional[RunStatus] = None,
    ) -> Run:
        """
        Updates a run with the new attributes values.
        :param run: The run name or id
        :param name: The name of the job
        :param system_name: The name of the system
        :param status: The status of the run
        :return: The updated Run
        """
        run_output = self._client.get_run(run, self._id)
        if name is not None and name != run_output.name:
            run_output["name"] = name
        if system_name is not None and system_name != run_output.system_name:
            run_output["systemName"] = system_name
        if status is not None and status != run_output.status:
            run_output["status"] = status
        updated_output = self._client.update_run(run_output)
        self._logger.info(f"Run with id: {updated_output.id} successfully updated.")
        return Run(
            updated_output.id,
            self,
            name=updated_output.name,
            system_name=updated_output.system_name,
            created_date=updated_output.created_date,
            start_date=updated_output.start_date,
            end_date=updated_output.end_date,
            status=updated_output.status,
        )

    def list_runs(
        self,
        search: Optional[str] = None,
        page_index: int = 1,
        page_size: int = 20,
    ) -> List[Run]:
        """
        Lists the runs in this job.

        :param search: The name to search
        :param page_index: The page index
        :param page_size: The page size

        :return: the list of Run of this jobs
        """
        outputs = self._client.list_runs(self.id, None, None, search, page_index, page_size)
        return [
            Run(
                run_output.id,
                self,
                run_output.name,
                run_output.system_name,
                run_output.created_date,
                run_output.start_date,
                run_output.end_date,
                run_output.status,
            )
            for run_output in outputs.list
        ]



from __future__ import annotations

import urllib
from typing import Optional, TYPE_CHECKING
from urllib.parse import urlencode

from ._auth import Auth
from ._http import HttpError
from .http_error_handlers import HttpErrorHandler
from .json import ModelOutput, ModelInput, Page, PagedResponse
from .project import ProjectApi
from .reference import InvalidReferenceError, MissingReferenceError

if TYPE_CHECKING:
    from vectice import Reference


class ModelApi:
    def __init__(self, auth: Auth):
        self._auth = auth
        self._httpErrorhandler = HttpErrorHandler()

    def get_model(
        self,
        model: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ):
        if not isinstance(model, int) and not isinstance(model, str):
            raise InvalidReferenceError("model", model)
        if isinstance(model, int):
            parent_project = None
            url = f"/metadata/model/{model}"
        else:
            if project is None:
                raise MissingReferenceError("project")
            parent_project = ProjectApi(self._auth).get_project(project, workspace)
            url = f"/metadata/project/{parent_project.id}/model/name/{urllib.parse.quote(model)}"
        try:
            response = self._auth._get(url)
            result = ModelOutput(response)
            if parent_project is not None:
                result.project = parent_project
            return result
        except HttpError as e:
            raise self._httpErrorhandler.handleGetHttpError(e, "model", model)

    def list_models(
        self,
        project: Reference,
        workspace: Optional[Reference] = None,
        search: Optional[str] = None,
        page_index=Page.index,
        page_size=Page.size,
    ) -> PagedResponse[ModelOutput]:
        queries = {"index": page_index, "size": page_size}
        if search:
            queries["search"] = search
        if isinstance(project, int):
            base_url = f"/metadata/project/{project}/model"
        else:
            parent_project = ProjectApi(self._auth).get_project(project, workspace)
            base_url = f"/metadata/project/{parent_project.id}/model"
        response = self._auth._get(f"{base_url}?{urlencode(queries)}")
        result = PagedResponse(
            item_cls=ModelOutput, total=response["total"], page=response["page"], items=response["items"]
        )
        return result

    def create_model(
        self,
        model: ModelInput,
        project: Reference,
        workspace: Optional[Reference] = None,
    ) -> ModelOutput:
        if model.get("name") is None:
            raise ValueError('"name" must be provided in model.')
        if isinstance(project, int):
            url = f"/metadata/project/{project}/model"
        elif isinstance(project, str):
            parent_project = ProjectApi(self._auth).get_project(project, workspace)
            url = f"/metadata/project/{parent_project.id}/model"
        else:
            raise InvalidReferenceError("project", project)
        try:
            response = self._auth._post(url, model)
        except HttpError as e:
            raise self._httpErrorhandler.handlePostHttpError(e, "model")
        return ModelOutput(**response)

    def update_model(
        self,
        data: ModelInput,
        model: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> ModelOutput:
        if isinstance(model, int):
            url = f"/metadata/model/{model}"
        elif isinstance(model, str):
            model_object = self.get_model(model, project, workspace)
            url = f"/metadata/model/{model_object.id}"
        else:
            raise InvalidReferenceError("model", model)
        try:
            response = self._auth._put(url, data)
            return ModelOutput(**response)
        except HttpError as e:
            raise self._httpErrorhandler.handlePostHttpError(e, "model")

    def delete_model(
        self,
        model: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ):
        try:
            model_object = self.get_model(model, project, workspace)
            self._auth._delete(f"/metadata/model/{model_object.id}")
        except HttpError as e:
            raise self._httpErrorhandler.handleGetHttpError(e, "model", model)



from __future__ import annotations


import urllib.parse

from typing import Optional, TYPE_CHECKING

from vectice.api._auth import Auth
from ._http import HttpError
from .http_error_handlers import HttpErrorHandler
from .json import WorkspaceOutput, PagedResponse
from .json.workspace import WorkspaceInput
from .reference import InvalidReferenceError

if TYPE_CHECKING:
    from vectice import Reference


class WorkspaceApi:
    def __init__(self, auth: Auth):
        self._auth = auth
        self._httpErrorhandler = HttpErrorHandler()

    def get_workspace(self, workspace: Reference) -> WorkspaceOutput:
        if isinstance(workspace, int):
            url = f"/metadata/workspace/{workspace}"
        elif isinstance(workspace, str):
            url = f"/metadata/workspace/name/{urllib.parse.quote(workspace)}"
        else:
            raise InvalidReferenceError("workspace", workspace)
        try:
            response = self._auth._get(url)
            return WorkspaceOutput(**response)
        except HttpError as e:
            raise self._httpErrorhandler.handleGetHttpError(e, "workspace", workspace)
        except TypeError:
            raise ValueError("The workspace is invalid. Please check the entered value.")

    def list_workspaces(
        self, search: Optional[str] = None, page_index: int = 1, page_size: int = 20
    ) -> PagedResponse[WorkspaceOutput]:
        if search is None or search == "":
            url = f"/metadata/workspace?index={page_index}&size={page_size}"
        else:
            url = f"/metadata/workspace?index={page_index}&size={page_size}&search={search}"
        workspaces = self._auth._get(url)
        return PagedResponse(
            item_cls=WorkspaceOutput, total=workspaces["total"], page=workspaces["page"], items=workspaces["items"]
        )

    def create_workspace(self, data: WorkspaceInput) -> WorkspaceOutput:
        url = "/metadata/workspace"
        try:
            response = self._auth._post(url, data.__dict__)
            return WorkspaceOutput(**response)
        except HttpError as e:
            raise self._httpErrorhandler.handlePostHttpError(e, "workspace")
        except TypeError:
            raise ValueError("The workspace is invalid. Please check the entered value.")

    def update_workspace(self, data: WorkspaceInput, workspace: Reference) -> WorkspaceOutput:
        if isinstance(workspace, int):
            url = f"/metadata/workspace/{workspace}"
        elif isinstance(workspace, str):
            workspace_object = self.get_workspace(workspace)
            url = f"/metadata/workspace/{workspace_object.id}"
        else:
            raise InvalidReferenceError("workspace", workspace)
        try:
            response = self._auth._put(url, data)
            return WorkspaceOutput(**response)
        except HttpError as e:
            raise self._httpErrorhandler.handlePutHttpError(e, "workspace", workspace)
        except TypeError:
            raise ValueError("The workspace is invalid. Please check the entered value.")



from __future__ import annotations

from typing import Optional, List, Union, TYPE_CHECKING

from vectice.api.json import JobArtifactType, ArtifactType

from .dataset_version import DatasetVersion
from .model_version import ModelVersion
from .code_version import CodeVersion
from vectice.api.reference import MissingReferenceError

if TYPE_CHECKING:
    from vectice.models import Dataset, Model


class Artifact:
    """
    Class for all type of artifact.

    This class links a Run to a Model version/Code version/Dataset version
    and allow to attach data to it (files/data).

    You can create artifact by using the factory class :class:`Artifacts`.
    You can create artifact by using the factory class :class:`Integrations`
    or simply using the :func:`~Run.add_artifact` method of a :class:`Run`.
    """

    def __init__(
        self,
        artifact: Optional[Union[Dataset, DatasetVersion, Model, ModelVersion, CodeVersion]] = None,
        type: Optional[JobArtifactType] = None,
        description: Optional[str] = None,
    ):
        self._artifact = artifact
        self._jobArtifactType = type
        self._description = description

    @property
    def artifactType(self) -> ArtifactType:
        """
        The artifact type: `MODEL`, `DATASET` or `CODEVERSION`.
        :return: Optional[JobArtifactType]
        """
        from vectice.models import Dataset, DatasetVersion, Model, ModelVersion, CodeVersion

        if isinstance(self._artifact, Model) or isinstance(self._artifact, ModelVersion):
            return ArtifactType.MODEL
        elif isinstance(self._artifact, Dataset) or isinstance(self._artifact, DatasetVersion):
            return ArtifactType.DATASET
        elif isinstance(self._artifact, CodeVersion):
            return ArtifactType.CODE
        else:
            raise RuntimeError(f"unexpected artifact type: {type(self._artifact)}")

    @property
    def artifact(self) -> Optional[Union[Dataset, DatasetVersion, Model, ModelVersion, CodeVersion]]:
        """
        The artifact object.
        :return: Optional[Union[Dataset, DatasetVersion, Model, ModelVersion, CodeVersion]]
        """
        return self._artifact

    @property
    def jobArtifactType(self) -> Optional[JobArtifactType]:
        """
        The artifact type for the job : `INPUT` or `OUTPUT`.
        :return: Optional[JobArtifactType]
        """
        return self._jobArtifactType

    @jobArtifactType.setter
    def jobArtifactType(self, value: JobArtifactType):
        self._jobArtifactType = value

    @property
    def description(self) -> Optional[str]:
        """
        The artifact description.
        :return: Optional[str]
        """
        return self._description

    @description.setter
    def description(self, description: Optional[str] = None):
        self._description = description

    def set_version(self, asset_version: Union[Dataset, DatasetVersion, Model, ModelVersion, CodeVersion]):
        """
        set the (Model or Dataset or Code) Version as the main content of the artifact.

        :param asset_version:
        """
        self._artifact = asset_version

    def add_attachments(self, file_path: Union[str, List[str]]) -> None:
        """
        Add some attachment/s to the artifact.

        The Attachment is directly uploaded to Vectice and can be viewed immediately if the operation.

        :param file_path: File path/s
        """
        if self.artifact is not None and (
            isinstance(self.artifact, DatasetVersion)
            or isinstance(self.artifact, ModelVersion)
            or isinstance(self.artifact, CodeVersion)
        ):
            self.artifact.add_attachments(file_path)
        else:
            raise MissingReferenceError("artifact")



from typing import Optional, List

from .artifact_version import ArtifactVersion
from .connection import ConnectionOutput
from .data_resource_schema import DataResourceSchema
from .data_resources_path import DataResourcePath
from .project import ProjectOutput


class DataResource:
    def __init__(
        self,
        name: str,
        description: Optional[str] = None,
        path: Optional[DataResourcePath] = None,
        schema: Optional[DataResourceSchema] = None,
        *args,
        **kwargs
    ):
        self._name = name
        self._description = description
        self._path = path
        self._schema = schema
        self._extra = kwargs

    @property
    def name(self) -> str:
        return self._name

    @property
    def description(self) -> Optional[str]:
        return self._description

    @property
    def path(self) -> Optional[DataResourcePath]:
        return self._path

    @property
    def schema(self) -> Optional[DataResourceSchema]:
        return self._schema


class DatasetInput(dict):
    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def pattern(self) -> str:
        return str(self["pattern"])

    @property
    def description(self) -> Optional[str]:
        if self.get("description", None):
            return str(self["description"])
        else:
            return None

    @property
    def connection_id(self) -> int:
        return int(self["connectionId"])

    @property
    def connection_name(self) -> str:
        return str(self["connectionName"])

    @property
    def version(self) -> ArtifactVersion:
        return ArtifactVersion(**self["version"])


class DatasetOutput(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if "project" in self:
            self._project: ProjectOutput = ProjectOutput(**self["project"])
        else:
            self._project = None
        if "connection" in self:
            self._connection: ConnectionOutput = ConnectionOutput(**self["connection"])
        else:
            self._connection = None
        self._resources: List[DataResource] = []
        if "dataResources" in self:
            for item in self["dataResources"]:
                path = item.pop("path", None)
                self._resources.append(DataResource(**item, path=DataResourcePath(**path)))

    def items(self):
        result = []
        for key in self:
            if self[key] is not None:
                result.append((key, self[key]))
        return result

    @property
    def id(self) -> int:
        return int(self["id"])

    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def description(self) -> str:
        return str(self["description"])

    @property
    def pattern(self) -> str:
        return str(self["pattern"])

    @property
    def is_pattern_base(self) -> str:
        return str(self["isPatternBase"])

    @property
    def create_date(self) -> str:
        return str(self["createdDate"])

    @property
    def updated_date(self) -> str:
        return str(self["updatedDate"])

    @property
    def deleted_date(self) -> str:
        return str(self["deletedDate"])

    @property
    def connection_id(self) -> Optional[int]:
        if self.get("connectionId", None):
            return int(self["connectionId"])
        else:
            return None

    @property
    def created_by_user_id(self) -> int:
        return int(self["createdByUserId"])

    @property
    def project_id(self) -> int:
        return int(self["projectId"])

    @property
    def version(self) -> int:
        return int(self["version"])

    @property
    def connection(self) -> Optional[ConnectionOutput]:
        return self._connection

    @property
    def project(self) -> ProjectOutput:
        return self._project

    @project.setter
    def project(self, project: ProjectOutput):
        self._project = project

    @property
    def resources(self) -> List[str]:
        return [item.path.uri for item in self._resources if item.path is not None and item.path.uri is not None]



from .artifact import Artifact

from .code import Code
from .code_version import CodeVersion
from .dataset import Dataset
from .dataset_version import DatasetVersion
from .errors import VecticeError
from .git_version import GitVersion
from .model import Model
from .model_version import ModelVersion
from .job import Job
from .run import Run
from .metric import Metric
from .property import Property
from .tag import Tag
from .user_version import UserVersion
from .data_resource import DataResource
from .api_token import ApiToken
from .workspace import Workspace
from .project import Project
from .connection import Connection
from .artifact_reference import ArtifactReference
from .attachment_container import AttachmentContainer
from .stage import Stage
from .block import Block

__all__ = [
    "ApiToken",
    "Artifact",
    "ArtifactReference",
    "AttachmentContainer",
    "Code",
    "CodeVersion",
    "Connection",
    "DataResource",
    "Dataset",
    "DatasetVersion",
    "GitVersion",
    "Job",
    "Metric",
    "Model",
    "ModelVersion",
    "Property",
    "Project",
    "Run",
    "Tag",
    "UserVersion",
    "VecticeError",
    "Workspace",
    "Stage",
    "Block",
]



from __future__ import annotations

import logging
import os.path
from typing import Optional, List, Union, Dict, BinaryIO, Any, TYPE_CHECKING

from vectice.api import BadReferenceError, Client
from vectice.api.json import (
    ModelType,
    JobType,
    ModelVersionStatus,
    WorkspaceInput,
    RunStatus,
    JobArtifactType,
    ArtifactType,
    FileMetadata,
    Page,
    StageStatus,
    VersionStrategy,
)
from vectice.api.reference import Reference, MissingReferenceError, InvalidReferenceError
from vectice.integrations import IntegrationFactory
from vectice.models import (
    Workspace,
    Project,
    Job,
    ModelVersion,
    Run,
    Dataset,
    DatasetVersion,
    Connection,
    Model,
    CodeVersion,
    GitVersion,
    ArtifactReference,
    Stage,
)
from vectice.models.integration import AbstractIntegration
from .__version__ import __version__
from .api.json.connection_type import ConnectionType
from .models.dataset_metadata import DatasetMetadata

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from botocore.session import Session


class Vectice:
    """
    High level class to list jobs and runs but also save runs

    The ability to toggle the autocode feature is enabled with the autocode argument. If autocode=True,
    Vectice will capture code artifacts relative to the .git file where you are executing your code. More can be found
    in the documentation at doc.vectice.com
    """

    def __init__(
        self,
        workspace: Optional[Reference] = None,
        project: Optional[Reference] = None,
        user_token: Optional[str] = None,
        api_endpoint: Optional[str] = None,
        lib: Optional[object] = None,
        allow_self_signed_certificate: bool = True,
        auto_log: bool = False,
    ):
        self._active_runs: Dict[int, Run] = {}
        logging.getLogger("Client").propagate = True
        self._client = Client(workspace, project, user_token, api_endpoint, True, allow_self_signed_certificate)
        self._auto_document_run = False
        self._document_stage: Optional[str] = None
        self._document_stage = None
        self._document_run = False
        logging.getLogger("Client").propagate = False
        self._logger = logging.getLogger(self.__class__.__name__)
        compatibility = self._client.check_compatibility()
        if compatibility.status != "OK":
            if compatibility.status == "Error":
                print(f"compatibility error : {compatibility.message}")
                self._logger.error(f"compatibility error : {compatibility.message}")
                raise RuntimeError(f"compatibility error : {compatibility.message}")
            else:
                print(f"compatibility warning : {compatibility.message}")
                self._logger.warning(f"compatibility warning : {compatibility.message}")
        try:
            self._integration_client: Optional[AbstractIntegration] = IntegrationFactory.create_adapter(
                lib=lib, vectice_client=self._client, auto_log=auto_log
            )
        except RuntimeError as e:
            self._logger.debug(f"Runtime {e}")
            self._integration_client = None

    def __repr__(self) -> str:
        lib = self._integration_client
        return (
            "Vectice("
            + f"workspace={self._client.workspace.name if self._client.workspace else 'None'}, "
            + f"project={self._client.project.name if self._client.project else 'None'}, "
            + f"api_endpoint={self._client._auth.api_base_url}, "
            + f"lib={lib.lib_name() if lib is not None else 'None'})"
        )

    @property
    def version(self) -> str:
        return __version__

    @property
    def project(self) -> Optional[Project]:
        if self._client.project is not None and self._client.workspace is not None and self.workspace is not None:
            result = Project(
                self._client.project.id, self.workspace, self._client.project.name, self._client.project.description
            )
            return result
        else:
            return None

    @property
    def workspace(self) -> Optional[Workspace]:
        if self._client.workspace is not None:
            result = Workspace(
                self._client.workspace.id, self._client.workspace.name, self._client.workspace.description
            )
            result.__post_init__(self._client, self._integration_client)
            return result
        else:
            return None

    def _get_parent_project(self, project, workspace) -> Project:
        if project is None and self.project is not None:
            project = self.project.id
        else:
            if project is None:
                raise MissingReferenceError("project")
            if isinstance(project, str) and workspace is None and self.workspace is not None:
                workspace = self.workspace.id
        project_output = self._client.get_project(project, workspace)
        workspace_output = self._client.get_workspace(project_output.workspace_id)
        parent_workspace = Workspace(workspace_output.id, workspace_output.name, workspace_output.description)
        parent_workspace.__post_init__(self._client, self._integration_client)
        return Project(project_output.id, parent_workspace, project_output.name, project_output.description)

    def list_workspaces(
        self,
        search: Optional[str] = None,
        page_index: int = 1,
        page_size: int = 20,
    ) -> List[Workspace]:
        """
        List workspaces.

        Lists the workspaces the user have access to with it's user connection.

        :return: List of Workspaces
        """
        outputs = self._client.list_workspaces(search, page_index, page_size)
        results = [Workspace(id=output.id, name=output.name, description=output.description) for output in outputs.list]
        for workspace in results:
            workspace.__post_init__(self._client, self._integration_client)
        return results

    def update_workspace(
        self, workspace: Reference, name: Optional[str] = None, description: Optional[str] = None
    ) -> Workspace:
        """
        Update workspace

        :param workspace:
        :param name:
        :param description:
        :return: the updated workspace
        """
        output = self._client.update_workspace(WorkspaceInput(name, description), workspace)
        result = Workspace(id=output.id, name=output.name, description=output.description)
        result.__post_init__(self._client, self._integration_client)
        self._logger.info(f"Workspace with id: {result.id} successfully updated.")
        return result

    def get_workspace(self, workspace: Reference) -> Workspace:
        """


        Gets the workspace with the specified reference (workspace name or id).

        :param workspace: The workspace name or id the user wants to access

        :return: Workspace
        """
        output = self._client.get_workspace(workspace)
        result = Workspace(output.id, output.name, output.description)
        result.__post_init__(self._client, self._integration_client)
        self._logger.info(f"Workspace with id: {result.id} successfully retrieved.")
        return result

    def get_project(self, project: Optional[Reference] = None, workspace: Optional[Reference] = None) -> Project:
        """
        Gets a project instance with the specified project name or id. If the user have the project id,
        it is enough to get the project otherwise the workspace should also be specified.

        :param workspace: The workspace name or id the project belongs to
        :param project: The project name or id to get.

        :return: A Project
        """
        if project is None and self.project is not None:
            return self.project
        else:
            if project is None:
                raise MissingReferenceError("project")
            if isinstance(project, str) and workspace is None and self.workspace is not None:
                workspace = self.workspace.id
        project_output = self._client.get_project(project, workspace)
        workspace_output = self._client.get_workspace(project_output.workspace_id)
        parent_workspace = Workspace(workspace_output.id, workspace_output.name, workspace_output.description)
        parent_workspace.__post_init__(self._client, self._integration_client)
        self._logger.info(f"Project with id: {project_output.id} successfully retrieved.")
        return Project(project_output.id, parent_workspace, project_output.name, project_output.description)

    # def get_job(self, job: Reference, project: Optional[Reference] = None, workspace: Optional[Reference] = None):
    #     job_output = self._client.get_job(job, project, workspace)
    #     project_output = self._client.get_project(job_output.project_id)
    #     workspace_object = Workspace(
    #         project_output.workspace.id, project_output.workspace.name, project_output.workspace.description
    #     )
    #     workspace_object.__post_init__(self._client)
    #     project_object = Project(
    #         project_output.id,
    #         workspace_object,
    #         project_output.name,
    #         project_output.description,
    #     )
    #     return Job(
    #         job_output.name,
    #         job_output.id,
    #         project_object,
    #         job_output.description,
    #         job_output.type,
    #     )

    def get_model(self, model: Reference, project: Optional[Reference] = None, workspace: Optional[Reference] = None):
        """
        Gets a model from the specified project.

        :param model: The model name or id
        :param workspace: The workspace name or id the project belongs to
        :param project: The project name or id to get.

        :return: A Model
        """
        if isinstance(model, int):
            model_output = self._client.get_model(model)
            if model_output is None:
                raise InvalidReferenceError("version", model)
            project_object = self._get_parent_project(model_output.project_id, workspace)
        elif isinstance(model, str):
            if project is None and self.project is not None:
                project = self.project.id
            if isinstance(project, str) and workspace is None and self.workspace is not None:
                workspace = self.workspace.id
            if project is None:
                raise MissingReferenceError("model", "project")
            project_object = self._get_parent_project(project, workspace)
        else:
            raise MissingReferenceError("model", "model")
        return project_object.get_model(model)

    def get_dataset(
        self, dataset: Reference, project: Optional[Reference] = None, workspace: Optional[Reference] = None
    ) -> Dataset:
        """
        Get a dataset.

        Gets a specific dataset from the specified project.

        :param dataset: dataset name or id
        :param project: The project name or id to get.
        :param workspace: The workspace name or id the project belongs to

        :return: A Dataset
        """
        if isinstance(dataset, int):
            dataset_output = self._client.get_dataset(dataset)
            if dataset_output is None:
                raise InvalidReferenceError("version", dataset)
            project_object = self._get_parent_project(dataset_output.project_id, workspace)
        elif isinstance(dataset, str):
            if project is None and self.project is not None:
                project = self.project.id
            if isinstance(project, str) and workspace is None and self.workspace is not None:
                workspace = self.workspace.id
            if project is None:
                raise MissingReferenceError("dataset", "project")
            project_object = self._get_parent_project(project, workspace)
        else:
            raise MissingReferenceError("dataset", "model")
        return project_object.get_dataset(dataset)

    def update_project(
        self,
        project: Reference,
        workspace: Optional[Reference] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
    ) -> Project:
        """
        Updates the information of a project.

        :param workspace: The workspace name or id the project belongs to
        :param project: The project name or id to update.
        :param name: The name of the project to be created.
        :param description: The description of the project to be created.

        :return: Project
        """
        if project is None and self.project is not None:
            return self.project
        else:
            if project is None:
                raise MissingReferenceError("project")
            if isinstance(project, str) and self.workspace is not None:
                workspace = self.workspace.id
        project_output = self._client.get_project(project, workspace)
        workspace_output = self._client.get_workspace(project_output.workspace_id)
        parent_workspace = Workspace(workspace_output.id, workspace_output.name, workspace_output.description)
        parent_workspace.__post_init__(self._client, self._integration_client)
        return parent_workspace.update_project(project, name, description)

    def create_model_version(
        self,
        name: str,
        model: Reference,
        workspace: Optional[Reference] = None,
        project: Optional[Reference] = None,
        description: Optional[str] = None,
        status: Optional[ModelVersionStatus] = ModelVersionStatus.EXPERIMENTATION,
        algorithm: Optional[str] = None,
        is_starred: Optional[bool] = False,
        metrics: Optional[Dict] = None,
        hyper_parameters: Optional[Dict] = None,
        attachments: Optional[List[str]] = None,
    ) -> ModelVersion:
        """
        Create a new model version.

        Creates a new model version with all the existing params in Vectice.

        :param model:
        :param name: The name of the model
        :param workspace: The workspace name or id the project belongs to
        :param project: The project name or id to get.
        :param description: The description of the model
        :param algorithm: The algorithm the model version uses
        :param is_starred: Whether the model is starred or not
        :param metrics: The model version metrics
        :param hyper_parameters: The model version hyper parameters

        :return: A ModelVersion
        """
        logging.getLogger("Project").propagate = False
        model_object = self.get_model(model, project, workspace)
        logging.getLogger("Project").propagate = True
        version: ModelVersion = model_object.create_model_version(
            name,
            description,
            status,
            algorithm,
            is_starred,
            metrics,
            hyper_parameters,
            None,
        )
        if attachments is not None:
            version.add_attachments(attachments)
        return version

    def create_gcs_dataset(
        self,
        uri: Union[str, List[str]],
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        service_account_json_path: Optional[str] = None,
    ) -> Optional[DatasetVersion]:
        """
        Create a Google Cloud Storage dataset version using an uri. Creates a Google Cloud Storage resource attached to the dataset version.
        The uri is found in the google cloud console in the Google Cloud Storage tab.

        Example:
        File level :  bucket_name/folder_name/file_name.csv (Attach the file)
        Folder level: bucket_name/folder_name (Attaches everything in the the folder)
        List : ['bucket_name/folder_name/file_name.csv', 'bucket_name/folder_name']

        :param workspace: The workspace name or id the project belongs to
        :param project: The project name or id to update.
        :param uri: The uri list of the Google Cloud Storage resources that will be used to create a dataset version
        :param name: The dataset name
        :param description: The description of the dataset version
        :return: A DatasetMetadataArtifact that can be used in runs as an input or output
        """

        def normalize_name(text: str) -> str:
            text = text.split("/")[-1]
            return text

        if name is None:
            if isinstance(uri, str):
                name = normalize_name(uri)
            else:
                name = normalize_name(os.path.commonprefix(uri))

        if project is None and self.project is not None:
            project = self.project.id
        else:
            if workspace is None and isinstance(project, str) and self.workspace is not None:
                workspace = self.workspace.id
        if project is None:
            raise MissingReferenceError("project", project)
        parent_project = self._get_parent_project(project, workspace)
        try:
            dataset = parent_project.get_dataset(name)
        except BadReferenceError:
            dataset = parent_project.create_dataset(
                name,
                description=description,
            )
        metadata = DatasetMetadata.create_gcs(uri, service_account_json_path)
        dataset_version = dataset.create_dataset_version(
            metadata=metadata, description=description, version_strategy=VersionStrategy.AUTOMATIC
        )
        return dataset_version

    def create_bigquery_dataset(
        self,
        uri: str,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        service_account_json_path: Optional[str] = None,
    ) -> Optional[DatasetVersion]:
        """
        Create a bigquery dataset version using an uri. Creates a BigQuery resource attached to the dataset version. The uri
        can be found in the table or dataset info.

        Example:
        Tables level : "Project_Name/Dataset_Name/Table_Name"
        Dataset level: "Project_Name/Dataset_Name"

        :param workspace: The workspace name or id the project belongs to
        :param project: The project name or id to update.
        :param uri: The url of the BigQuery dataset version that will be created
        :param name: The dataset name
        :param description: The description of the dataset version
        :return: A DatasetMetadataArtifact that can be used in runs as an input or output
        """

        def normalize_name(text: str) -> str:
            text = text.split("/")[-1]
            return text

        if name is None:
            if isinstance(uri, str):
                name = normalize_name(uri)
            else:
                name = normalize_name(os.path.commonprefix(uri))

        if project is None and self.project is not None:
            project = self.project.id
        else:
            if workspace is None and isinstance(project, str) and self.workspace is not None:
                workspace = self.workspace.id
        if project is None:
            raise MissingReferenceError("project", project)
        parent_project = self._get_parent_project(project, workspace)
        try:
            dataset = parent_project.get_dataset(name)
        except BadReferenceError:
            dataset = parent_project.create_dataset(
                name,
                description=description,
            )
        metadata = DatasetMetadata.create_bigquery(uri, service_account_json_path)
        dataset_version = dataset.create_dataset_version(
            metadata=metadata, description=description, version_strategy=VersionStrategy.AUTOMATIC
        )
        return dataset_version

    def create_s3_dataset(
        self,
        uri: str,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
        aws_session_token: Optional[str] = None,
        profile_name: Optional[str] = None,
        region_name: Optional[str] = None,
        botocore_session: Optional[Session] = None,
    ) -> Optional[DatasetVersion]:
        """
        For credentials, please read AWS instructions at https://boto3.amazonaws.com/v1/documentation/api/latest/guide/credentials.html
        As describe the AWS client also support environment variables `AWS_ACCESS_KEY_ID `, `AWS_SECRET_ACCESS_KEY` and `AWS_SESSION_TOKEN`.
        It also support credentials file in `~/.aws/credentials`.

        :param uri: list of files/folders to be used for the dataset. the metadata will be extracted from this list.
        :type aws_access_key_id: string
        :param aws_access_key_id: AWS access key ID
        :type aws_secret_access_key: string
        :param aws_secret_access_key: AWS secret access key
        :type aws_session_token: string
        :param aws_session_token: AWS temporary session token
        :type region_name: string
        :param region_name: Default region when creating new connections
        :type profile_name: string
        :param profile_name: The name of a profile to use. If not given, then
                             the default profile is used.
        :type botocore_session: botocore.session.Session
        :param botocore_session: Use this Botocore session instead of creating
                                 a new default one.
        :param uri:
        :param project:
        :param workspace:
        :param name:
        :param description:
        :return:
        """

        def normalize_name(text: str) -> str:
            text = text.split("/")[-1]
            return text

        if name is None:
            if isinstance(uri, str):
                name = normalize_name(uri)
            else:
                name = normalize_name(os.path.commonprefix(uri))

        if project is None and self.project is not None:
            project = self.project.id
        else:
            if workspace is None and isinstance(project, str) and self.workspace is not None:
                workspace = self.workspace.id
        if project is None:
            raise MissingReferenceError("project", project)
        parent_project = self._get_parent_project(project, workspace)
        try:
            dataset = parent_project.get_dataset(name)
        except BadReferenceError:
            dataset = parent_project.create_dataset(
                name,
                description=description,
            )
        metadata = DatasetMetadata.create_s3(
            uri,
            aws_access_key_id,
            aws_secret_access_key,
            aws_session_token,
            profile_name,
            region_name,
            botocore_session,
        )
        dataset_version = dataset.create_dataset_version(
            metadata=metadata, description=description, version_strategy=VersionStrategy.AUTOMATIC
        )
        return dataset_version

    def create_project(
        self, name: str, workspace: Optional[Reference] = None, description: Optional[str] = None
    ) -> Project:
        """
        Creates a new project in a workspace with the specified name. If no workspace name is given the project is created in the default workspace.

        :param workspace: The workspace name or id in which the project should be created.
        :param name: The name of the project to be created.
        :param description: The description of the project to be created.

        :return: Project
        """
        if workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        if workspace is None:
            raise MissingReferenceError("workspace", workspace)
        workspace_output = self._client.get_workspace(workspace)
        parent_workspace = Workspace(workspace_output.id, workspace_output.name, workspace_output.description)
        parent_workspace.__post_init__(self._client, self._integration_client)
        return parent_workspace.create_project(name, description)

    def delete_project(
        self,
        project: Reference,
        workspace: Optional[Reference] = None,
    ) -> None:
        """
        Deletes the project specified by the user.

        :param workspace: The workspace name or id the project belongs to
        :param project: The project name or id to delete.

        :return: None
        """
        if workspace is None and self.workspace is not None:
            workspace = self.workspace.id

        if workspace is not None:
            workspace_output = self._client.get_workspace(workspace)
            parent_workspace = Workspace(workspace_output.id, workspace_output.name, workspace_output.description)
            parent_workspace.__post_init__(self._client, self._integration_client)
            parent_workspace.delete_project(project)
        elif isinstance(project, int):
            project_instance = self._get_parent_project(project, None)
            project_instance.delete()  # This line is not working
        else:
            raise InvalidReferenceError("project", project)

    def list_projects(
        self,
        workspace: Optional[Reference] = None,
        search: Optional[str] = None,
        page_index: int = 1,
        page_size: int = 20,
    ) -> List[Project]:
        """
        Lists the projects available in the specified workspace. If No workspace reference(name or id) is given, the default workspace is selected.

        :param workspace: The workspace name or id the projects belong to
        :param search: The name to search
        :param page_index: The page index
        :param page_size: The page size
        :return: List of Projects
        """
        workspace_object = self.workspace if workspace is None else self.get_workspace(workspace)
        if workspace_object is None:
            raise InvalidReferenceError("workspace", workspace)

        return workspace_object.list_projects(search, page_index, page_size)

    def create_dataset(
        self,
        name: str,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        pattern: str = "*",
        description: Optional[str] = None,
        connection: Optional[Reference] = None,
        resources: Optional[List[str]] = None,
    ) -> Dataset:
        """
        Create a new dataset

        Creates a new dataset in a specific project. File/s and folder/s create resource/s associated with a dataset version
        that will belong to the created dataset.

        :param name: The name of the dataset
        :param project: The project name or id
        :param workspace: The workspace name or id
        :param pattern: The dataset pattern
        :param description: The description of the dataset
        :param properties: The properties of the dataset version
        :param connection: The connection name or id the dataset belong to
        :param resources: The file/s of the dataset version

        :return: Dataset
        """
        if project is None and self.project is not None:
            project = self.project.id
            workspace = None
        elif isinstance(project, str) and workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        parent_project = self._get_parent_project(project, workspace)
        return parent_project.create_dataset(name, pattern, description, connection, resources)

    def create_job(
        self,
        name: str,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        description: Optional[str] = None,
        type: JobType = JobType.OTHER,
    ) -> Job:
        """
        Create a new job in the project with the given attributes.

        The name is required and must be unique in the project.

        The default job type is `OTHER`. check the :class:`JobType` for other possible values.

        :param workspace: The workspace name or id the project belongs to
        :param project: The project name or id the job should belong to.
        :param name: name of the job
        :param description: quick description of the job
        :param type: the type of the job
        :return: the newly created job
        """
        if project is None and self.project is not None:
            project = self.project.id
            workspace = None
        elif isinstance(project, str) and workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        if project is None:
            raise MissingReferenceError("job", "project")
        parent_project = self._get_parent_project(project, workspace)
        return parent_project.create_job(name, description, type)

    def list_jobs(
        self,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        search: Optional[str] = None,
        page_index: int = 1,
        page_size: int = 20,
    ) -> List[Job]:
        """
        List the jobs.

        Lists the jobs in a specified project. The search parm helps filter the jobs depending on the name given.

        :param workspace: The workspace name or id the project belongs to
        :param project: The project name or id to get.
        :param search: The name to search
        :param page_index: The page index
        :param page_size: The page size

        :return: List of Jobs
        """
        if project is None and self.project is not None:
            project = self.project.id
            workspace = None
        elif isinstance(project, str) and workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        if project is None:
            raise MissingReferenceError("job", "project")
        parent_project = self._get_parent_project(project, workspace)
        return parent_project.list_jobs(search, page_index, page_size)

    def delete_job(
        self,
        job: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> None:
        """
        Deletes the specified job in a specific project.

        :param workspace: The workspace name or id the project belongs to
        :param project: The project name or id to get.
        :param job: The job name or id

        :return: None
        """
        logging.getLogger("Project").propagate = False
        if isinstance(job, int):
            job_output = self._client.get_job(job)
            parent_project = self._get_parent_project(job_output.project_id, workspace)
            logging.getLogger("Project").propagate = True
            parent_project.delete_job(job_output.id)
        elif isinstance(job, str):
            if project is None and self.project is not None:
                project = self.project.id
            if project is None:
                raise MissingReferenceError("job", "project")
            parent_project = self._get_parent_project(project, workspace)
            job_output = parent_project._client.get_job(job)
            logging.getLogger("Project").propagate = True
            parent_project.delete_job(job_output.id)
        else:
            raise InvalidReferenceError("job", job)

    def get_job(
        self,
        job: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> Job:
        """
        Gets the specified job in a specific project.

        :param workspace: The workspace name or id the project belongs to
        :param project: The project name or id to get.
        :param job: The job name or id

        :return: A Job
        """
        logging.getLogger("Project").propagate = False
        if isinstance(job, int):
            job_output = self._client.get_job(job)
            parent_project = self._get_parent_project(job_output.project_id, workspace)
        else:
            if project is None and self.project is not None:
                project = self.project.id
                workspace = None
            elif isinstance(project, str) and workspace is None and self.workspace is not None:
                workspace = self.workspace.id
            if project is None:
                raise MissingReferenceError("job", "project")
            parent_project = self._get_parent_project(project, workspace)
        logging.getLogger("Project").propagate = True
        return parent_project.get_job(job)

    def update_job(
        self,
        job: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        name: Optional[str] = None,
        job_type: Optional[JobType] = None,
        description: Optional[str] = None,
    ) -> Job:
        """
        Updates the specified job with the new information.

        :param workspace: The workspace name or id the project belongs to
        :param project: The project name or id to get.
        :param job: The job name or id
        :param name: The name of the job
        :param job_type: The job type
        :param description: The description of the job

        :return: None
        """
        logging.getLogger("Project").propagate = False
        if isinstance(job, int):
            job_output = self._client.get_job(job)
            parent_project = self._get_parent_project(job_output.project.id, job_output.project.workspace.id)
            job_object = Job(job_output.name, job_output.id, parent_project, job_output.description, job_output.type)
        elif isinstance(job, str):
            if project is None and self.project is not None:
                project = self.project.id
            if project is None:
                raise MissingReferenceError("job", "project")
            if workspace is None and isinstance(project, str) and self.workspace is not None:
                workspace = self.workspace.id
            job_output = self._client.get_job(job, project, workspace)
            parent_project = self._get_parent_project(job_output.project.id, job_output.project.workspace.id)
            job_object = Job(job_output.name, job_output.id, parent_project, job_output.description, job_output.type)
        else:
            raise InvalidReferenceError("job", job)
        logging.getLogger("Project").propagate = True
        return job_object.project.update_job(job, name, description, job_type)

    def start_run(
        self,
        run: Run,
        inputs: Optional[List[Union[ArtifactReference, CodeVersion, ModelVersion, DatasetVersion]]] = None,
    ) -> Run:
        """
        Start the run created before by calling :func:`~Vectice.create_run` function

        :param run: The runnable job to start
        :param inputs: A list of artifacts used as inputs by this run.
        :return: A reference to a run in progress
        """
        run.start(inputs)
        self._active_runs[run.id] = run
        return run

    def end_run(
        self,
        run: Run,
        outputs: Optional[List[Union[ArtifactReference, CodeVersion, ModelVersion, DatasetVersion]]] = None,
        status: RunStatus = RunStatus.COMPLETED,
    ) -> Optional[int]:
        """
        End the current (last) active run started by :func:`~Vectice.start_run`.
        To end a specific run, use :func:`~Vectice.stop_run` instead.

        :return: Identifier of the run in Vectice if successfully saved
        """
        run.end_run(outputs, status)
        if self._auto_document_run:
            """
            Captures an existing run. That is still running.
            """
            if self._document_stage is not None:
                try:
                    stage = self.get_stage(self._document_stage)
                except Exception:
                    logger.debug(f"Stage {self._document_stage} for auto documentation not found.")
                    stage = self.create_stage(self._document_stage, status=StageStatus.InProgress)
            else:
                stage = self.create_stage(f"Auto Document {run.name}", status=StageStatus.InProgress)
            run._capture_run_info(stage)
        del self._active_runs[run.id]
        return run.id

    def create_run(
        self,
        job: Reference,
        workspace: Optional[Reference] = None,
        project: Optional[Reference] = None,
        name: Optional[str] = None,
        properties: Optional[Dict[str, Any]] = None,
        auto_code: bool = False,
        check_remote_repository: bool = True,
    ) -> Run:
        """
        Create a run.

        Creates a run in the specified project with the information passed to this method.

        The run is not viewable in the Vectice platform until the job is started.

        See :class:`Run` for more information

        :param job: The job name or id the run belongs to
        :param workspace: The workspace name or id the run belongs to
        :param project: The project name or id the run belongs to
        :param name: The run name.
        :param properties: Properties of the run
        :param auto_code: If the run has auto code enabled
        :param check_remote_repository: If the run has the auto code, remote repository check enabled


        :return: Run
        """
        if project is None and self.project is not None:
            project = self.project.id
            workspace = None
        elif isinstance(project, str) and workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        if project is None:
            raise MissingReferenceError("job", "project")
        job_output = self._client.get_job(job, project, workspace)
        parent_project = self._get_parent_project(job_output.project.id, job_output.project.workspace.id)
        parent_job = Job(job_output.name, job_output.id, parent_project, job_output.description, job_output.type)
        run = parent_job.create_run(name, properties, auto_code, check_remote_repository)
        return run

    def list_runs(
        self,
        job: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        search: Optional[str] = None,
        page_index: int = 1,
        page_size: int = 20,
    ) -> List[Run]:
        """
        List the runs.

        Lists the runs in a specific project.

        :param workspace: The workspace name or id the project belongs to
        :param project: The project name or id to get.
        :param job: The job name or id
        :param page_index: The page index
        :param page_size: The page size

        :return: List of JobRuns
        """
        if project is None and self.project is not None:
            project = self.project.id
            workspace = None
        elif isinstance(project, str) and workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        if project is None:
            raise MissingReferenceError("job", "project")
        job_output = self._client.get_job(job, project, workspace)
        parent_project = self._get_parent_project(job_output.project.id, job_output.project.workspace.id)
        parent_job = Job(job_output.name, job_output.id, parent_project, job_output.description, job_output.type)
        logging.getLogger("Project").propagate = True
        return parent_job.list_runs(search, page_index, page_size)

    def delete_run(
        self,
        run: Reference,
        job: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> None:
        """
        Deletes the specified run in a specific project.

        :param workspace: The workspace name or id the project belongs to
        :param project: The project name or id to get.
        :param run: The run name or id
        :param job: The job name or id

        :return: None
        """
        logging.getLogger("Vectice").propagate = False
        logging.getLogger("Project").propagate = False
        if isinstance(run, int):
            run_object = self.get_run(run, job)
        elif isinstance(run, str):
            if job is None:
                raise MissingReferenceError("run", "job")
            if project is None and self.project is not None:
                project = self.project.id
            if project is None:
                raise MissingReferenceError("job", "project")
            if workspace is None and isinstance(project, str) and self.workspace is not None:
                workspace = self.workspace.id
            run_object = self.get_run(run, job, project, workspace)
        else:
            raise InvalidReferenceError("job", job)
        logging.getLogger("Project").propagate = True
        logging.getLogger("Vectice").propagate = True
        run_object.job.delete_run(run)

    def get_run(
        self,
        run: Reference,
        job: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> Run:
        """
        Get a Run in the specified workspace/project/job.

        :param workspace: The workspace name or id the project belongs to
        :param project: The project name or id to get.
        :param job: The job name or id
        :param run: The run name or id

        :return: A Run
        """
        if isinstance(run, int):
            run_output = self._client.get_run(run, job, project, workspace)
            job_output = self._client.get_job(run_output.job_id, project, workspace)
            parent_project = self._get_parent_project(job_output.project.id, job_output.project.workspace_id)
            job_object = Job(job_output.name, job_output.id, parent_project, job_output.description, job_output.type)
            self._logger.info(f"Run with id: {run_output.id} successfully retrieved.")
            result = Run(
                id=run_output.id,
                job=job_object,
                name=run_output.name,
                system_name=run_output.system_name,
                created_date=run_output.created_date,
                start_date=run_output.start_date,
                end_date=run_output.end_date,
                status=run_output.status,
                description=run_output.description,
            )
        else:
            if job is None:
                raise MissingReferenceError("run", "job")
            if project is None and self.project is not None:
                project = self.project.id
            if project is None:
                raise MissingReferenceError("job", "project")
            if workspace is None and isinstance(project, str) and self.workspace is not None:
                workspace = self.workspace.id
            job_output = self._client.get_job(job, project, workspace)
            parent_project = self._get_parent_project(job_output.project.id, job_output.project.workspace_id)
            job_object = Job(job_output.name, job_output.id, parent_project, job_output.description, job_output.type)
            logging.getLogger("Job").propagate = False
            result = job_object.get_run(run)
            logging.getLogger("Job").propagate = True
        artifacts = self._client.list_artifacts(result.id).list
        for artifact in artifacts:

            if artifact.artifact_type == ArtifactType.DATASET:
                logging.getLogger("Dataset").propagate = False
                artifact_version: Union[DatasetVersion, ModelVersion, CodeVersion] = self.get_dataset_version(
                    artifact.artifact_id
                )
                logging.getLogger("Dataset").propagate = True
            elif artifact.artifact_type == ArtifactType.MODEL:
                logging.getLogger("Model").propagate = False
                artifact_version = self.get_model_version(artifact.artifact_id)
                logging.getLogger("Model").propagate = False
            else:
                logging.getLogger("Project").propagate = False
                artifact_version = self.get_code_version(artifact.artifact_id)
                logging.getLogger("Project").propagate = False
            reference = ArtifactReference(
                version_id=artifact.artifact_id,
                version_name=artifact_version.name,
                version_number=artifact_version.version_number,
                dataset=artifact_version.dataset.name if isinstance(artifact_version, DatasetVersion) else None,
                model=artifact_version.model.name if isinstance(artifact_version, ModelVersion) else None,
                code=artifact_version.id if isinstance(artifact_version, CodeVersion) else None,
                description=artifact_version.description,
            )
            if artifact.job_artifact_type == JobArtifactType.INPUT:
                result._inputs.append(reference)
            else:
                result._outputs.append(reference)
        logging.getLogger("Job").propagate = True
        return result

    def update_run(
        self,
        run: Reference,
        job: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        name: Optional[str] = None,
        system_name: Optional[str] = None,
        status: Optional[RunStatus] = None,
    ) -> Run:
        """
        Update a run.

        Updates the specified run with the new information.

        :param workspace: The workspace name or id the project belongs to
        :param project: The project name or id to get.
        :param job: The job name or id
        :param run: The run name or id
        :param name: The name of the job
        :param system_name: The name of the system
        :param status: The status of the run

        :return: the updated Run
        """
        logging.getLogger("Project").propagate = False
        logging.getLogger("Dataset").propagate = False
        logging.getLogger("Vectice").propagate = False
        if isinstance(run, int):
            run_object = self.get_run(run)
        elif isinstance(run, str):
            if job is None:
                raise MissingReferenceError("run", "job")
            if project is None and self.project is not None:
                project = self.project.id
            if project is None:
                raise MissingReferenceError("job", "project")
            if workspace is None and isinstance(project, str) and self.workspace is not None:
                workspace = self.workspace.id
            run_object = self.get_run(run, job, project, workspace)
        else:
            raise InvalidReferenceError("job", job)
        logging.getLogger("Project").propagate = True
        logging.getLogger("Dataset").propagate = True
        logging.getLogger("Vectice").propagate = True
        result: Run = run_object.job.update_run(run, name, system_name, None if status is None else status)
        return result

    def add_model_version_attachment(
        self,
        file: str,
        model_version: Reference,
        model: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        project: Optional[Reference] = None,
    ) -> None:
        """
        Attach a file to an existing Model Version.

        The Model Version Attachment will be added and associated to the Model Version if it exists.
        The Attachment is directly uploaded to Vectice and can be viewed immediately if the operation
        is successful. The Attachment can be seen in the Vectice UI or interacted with with the Python SDK.

        :param file: A file to attach on the model version
        :param model: The model name or id
        :param model_version: The model version name or id
        :param workspace: The workspace name or id
        :param project: The project name or id
        :return: None
        """
        model_version_object = self.get_model_version(model_version, model, project, workspace)
        model_version_object.add_attachments(file)

    def delete_model_version_attachment(
        self,
        file: str,
        model_version: Reference,
        model: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        project: Optional[Reference] = None,
    ) -> None:
        """
        Delete a Model Version Attachment.

        Deletes the specified attachment from the model version.

        :param model: The model name or id
        :param model_version: The model version name or id
        :param file: The file name
        :param workspace: The workspace name or id
        :param project: The project name or id
        :return: None
        """
        model_version_object = self.get_model_version(model_version, model, project, workspace)
        model_version_object.delete_attachments(file)

    def get_model_version_attachment(
        self,
        file: str,
        model_version: Reference,
        model: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        project: Optional[Reference] = None,
    ) -> Optional[BinaryIO]:
        """
        Get a model version attachment.

        Gets the attachment of a model version.

        :param model: The model name or id
        :param model_version: The model version name or id
        :param file: The file name
        :param workspace: The workspace name or id
        :param project: The project name or id
        :return: An Artifact Version Attachment
        """
        model_version_object = self.get_model_version(model_version, model, project, workspace)
        return model_version_object.get_attachment(file)

    def list_model_version_attachments(
        self,
        model_version: Reference,
        model: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        project: Optional[Reference] = None,
    ) -> List[str]:
        """
        List the files attached to a Model version.

        :param model: The model name or id
        :param model_version: The model version name or id
        :param workspace: The workspace name or id
        :param project: The project name or id
        :return: A list of file names
        """
        model_version_object = self.get_model_version(model_version, model, project, workspace)
        result = model_version_object.list_attachments()
        return result

    def add_dataset_version_attachment(
        self,
        file: str,
        dataset_version: Reference,
        dataset: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        project: Optional[Reference] = None,
    ) -> None:
        """
        Adds a Dataset Version Attachment to an existing Dataset Version.

        The Dataset Version Attachment will be added and associated to the Dataset Version if it exists.
        The Attachment is directly uploaded to Vectice and can be viewed immediately if the operation
        is successful. The Attachment can be seen in the Vectice UI or interacted with with the Python SDK.

        :param workspace: The workspace name or id
        :param project: The project name or id
        :param file: A file to attach on the dataset version
        :param dataset: The dataset name or id
        :param dataset_version: The dataset version name or id

        :return: None
        """
        dataset_version_object: DatasetVersion = self.get_dataset_version(dataset_version, dataset, project, workspace)
        dataset_version_object.add_attachments(file)

    def delete_dataset_version_attachment(
        self,
        file: str,
        dataset_version: Reference,
        dataset: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        project: Optional[Reference] = None,
    ) -> None:
        """
        Deletes the specified attachment from the dataset version.

        :param workspace: The workspace name or id
        :param project: The project name or id
        :param dataset: The dataset name or id
        :param dataset_version: The dataset version name or id
        :param file: The file name

        :return: None
        """
        dataset_version_object: DatasetVersion = self.get_dataset_version(dataset_version, dataset, project, workspace)
        dataset_version_object.delete_attachments(file)

    def list_dataset_version_attachments(
        self,
        dataset_version: Reference,
        dataset: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> List[str]:
        """
        List the files attached to a Dataset version.
        :param workspace: The workspace name or id
        :param project: The project name or id
        :param dataset: The dataset name or id
        :param dataset_version: The dataset version name or id

        :return: An Artifact Version Attachment
        """
        dataset_version_object: DatasetVersion = self.get_dataset_version(dataset_version, dataset, project, workspace)
        return dataset_version_object.list_attachments()

    def add_code_version_attachment(
        self,
        file: Union[str, List[str]],
        version: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> None:
        """
        Adds a Code Version Attachment to an existing Code Version.

        The Code Version Attachment will be added and associated to the Code Version if it exists.
        The Attachment is directly uploaded to Vectice and can be viewed immediately if the operation
        is successful. The Attachment can be seen in the Vectice UI or interacted with with the Python SDK.

        :param workspace: The workspace name or id
        :param project: The project name or id
        :param file: A file or list of files to attach on the dataset version
        :param code_version: The code version name or id

        :return: None
        """

        code_version = self.get_code_version(version, project, workspace)
        if code_version is None:
            raise ValueError("The code version was not found. Please check the reference provided.")
        code_version.add_attachments(file)

    def delete_code_version_attachment(
        self,
        file: Union[str, List[str]],
        version: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> None:
        """
        Deletes the specified attachment from the code version.

        :param workspace: The workspace name or id
        :param project: The project name or id
        :param code_version: The code version name or id
        :param file: The file name

        :return: None
        """
        code_version = self.get_code_version(version, project, workspace)
        code_version.delete_attachments(file)

    def get_code_version_attachment(
        self,
        file: str,
        version: Reference,
        workspace: Optional[Reference] = None,
        project: Optional[Reference] = None,
    ) -> Optional[BinaryIO]:
        """
        Gets the attachment of a code version.

        :param workspace: The workspace name or id
        :param project: The project name or id
        :param code_version: The code version name or id
        :param file: The file name

        :return: A list of file name or uri
        """
        code_version = self.get_code_version(version, project, workspace)
        return code_version.get_attachment(file)

    def list_code_version_attachments(
        self,
        version: Reference,
        workspace: Optional[Reference] = None,
        project: Optional[Reference] = None,
    ) -> List[str]:
        """
        List the files attached to a Code version.


        :param workspace: The workspace name or id
        :param project: The project name or id
        :param code_version: The code version name or id
        :return: A list of file name or uri
        """
        code_version = self.get_code_version(version, project, workspace)
        return code_version.list_attachments()

    def add_run_attachment(
        self,
        file: str,
        run: Reference,
        job: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        project: Optional[Reference] = None,
    ) -> None:
        """
        Adds a Code Version Attachment to an existing Code Version.

        The Code Version Attachment will be added and associated to the Code Version if it exists.
        The Attachment is directly uploaded to Vectice and can be viewed immediately if the operation
        is successful. The Attachment can be seen in the Vectice UI or interacted with with the Python SDK.

        :param workspace: The workspace name or id
        :param project: The project name or id
        :param file: A file to attach on the dataset version
        :param run: The run name or id
        :param job: The job name or id
        :return: None
        """
        run_object: Run = self.get_run(run, job, project, workspace)
        run_object.add_attachments(file)

    def delete_run_attachment(
        self,
        file: str,
        run: Reference,
        job: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        project: Optional[Reference] = None,
    ) -> None:
        """
        Delete a Code Version Attachment.
        Deletes the specified attachment from the code version.

        :param workspace: The workspace name or id
        :param project: The project name or id
        :param run: The run name or id
        :param job: The job name or id
        :param file: The file name

        :return: None
        """
        run_object = self.get_run(run, job, project, workspace)
        run_object.delete_attachments(file)

    def get_run_attachment(
        self,
        file: str,
        run: Reference,
        job: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> Optional[BinaryIO]:
        """
        Gets the attachment of a code version.

        :param workspace: The workspace name or id
        :param project: The project name or id
        :param run: The run name or id
        :param job: The job name or id
        :param file: The file path

        :return: A list of file name or uri
        """
        run_object: Run = self.get_run(run, job, project, workspace)
        return run_object.get_attachment(file)

    def list_run_attachments(
        self,
        run: Reference,
        job: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> List[str]:
        """
        List the files attached to a Code version.

        :param workspace: The workspace name or id
        :param project: The project name or id
        :param run: The run name or id
        :param job: The job name or id
        :return: A list of file name or uri
        """
        run_object: Run = self.get_run(run, job, project, workspace)
        return run_object.list_attachments()

    def create_dataset_version(
        self,
        dataset: Reference,
        name: Optional[str] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        description: Optional[str] = None,
        is_starred: Optional[bool] = None,
        properties: Optional[Dict[str, Any]] = None,
        resources: Optional[List[str]] = None,
        metadata: Optional[List[FileMetadata]] = None,
        attachments: Optional[List[str]] = None,
        version_strategy: VersionStrategy = VersionStrategy.AUTOMATIC,
    ) -> DatasetVersion:
        """
        Create a dataset version.

        Creates a dataset version of the specified dataset. Resources are cloud provider assets e.g Google Cloud Storage
        with a connection in the Vectice UI and metadata are assets with no connection in Vectice UI.

        :param dataset: dataset name or id
        :param name: The name of the dataset version e.g Changes 'Version 1' to 'name'
        :param workspace: The workspace name or id the project belongs to
        :param project: The project name or id to get.
        :param description: The description of the dataset version
        :param is_starred: Whether the dataset version is starred
        :param properties: The properties of the dataset version
        :param resources: The resource uri/s of the dataset version
        :param metadata: The metadata uri/s of the dataset version


        :return: A DatasetVersion
        """
        if project is None and self.project is not None:
            project = self.project.id
            workspace = None
        elif isinstance(project, str) and workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        dataset_output = self._client.get_dataset(dataset, project, workspace)
        parent_project = self._get_parent_project(dataset_output.project_id, dataset_output.project.workspace_id)
        dataset_object: Dataset = Dataset(
            dataset_output.name,
            dataset_output.id,
            project=parent_project,
            description=dataset_output.description,
        )
        result = dataset_object.create_dataset_version(
            name, description, is_starred, properties, resources, metadata, attachments, version_strategy
        )
        return result

    def list_dataset_versions(
        self,
        dataset: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        search: Optional[str] = None,
        page_index: int = 1,
        page_size: int = 20,
    ) -> List[DatasetVersion]:
        """
        List the dataset versions.

        Lists the dataset versions of the specified dataset.

        :param dataset: dataset name or id
        :param workspace: The workspace name or id the project belongs to
        :param project: The project name or id to get.

        :return: List of DatasetVersions
        """
        parent_dataset = self.get_dataset(dataset, project, workspace)
        return parent_dataset.list_dataset_versions(search, page_index, page_size)

    def get_dataset_version(
        self,
        version: Reference,
        dataset: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> DatasetVersion:
        """
        Get a dataset version.

        Gets a specific dataset version of the specified dataset.

        :param dataset: dataset name or id
        :param version: The dataset version name or id
        :param workspace: The workspace name or id the project belongs to
        :param project: The project name or id to get.

        :return: A DatasetVersion
        """
        logging.getLogger("Project").propagate = False
        if isinstance(version, int):
            dataset_version_output = self._client.get_dataset_version(version)
            if dataset_version_output is None:
                raise InvalidReferenceError("dataset version", version)
            parent_dataset: Dataset | None = self.get_dataset(dataset_version_output.dataset_id)
        else:
            if dataset is None:
                raise MissingReferenceError("dataset version", "dataset")
            parent_dataset = self.get_dataset(dataset, project, workspace)
        if parent_dataset is None:
            raise InvalidReferenceError("dataset", dataset)
        logging.getLogger("Project").propagate = True
        return parent_dataset.get_dataset_version(version)

    def get_dataset_version_attachment(
        self,
        file: str,
        dataset_version: Reference,
        dataset: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        project: Optional[Reference] = None,
    ) -> Optional[BinaryIO]:
        """
        Get a model version attachment.

        Gets the attachment of a model version.

        :param dataset: The model name or id
        :param dataset_version: The model version name or id
        :param file: The file name
        :param workspace: The workspace name or id
        :param project: The project name or id
        :return: An Artifact Version Attachment
        """
        dataset_version_object = self.get_dataset_version(dataset_version, dataset, project, workspace)
        return dataset_version_object.get_attachment(file)

    def list_dataset_version_files_metadata(
        self,
        version: Reference,
        dataset: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> List[FileMetadata]:
        dataset_version = self.get_dataset_version(version, dataset, project, workspace)
        return dataset_version.list_files_metadata()

    def list_datasets(
        self,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        search: Optional[str] = None,
        page_index: int = Page.index,
        page_size: int = Page.size,
    ) -> List[Dataset]:
        """
        List the datasets in a workspace.

        List the users connections having access to the specified workspace.

        :param workspace: The workspace name or id
        :param search: The connection name to search

        :return: List of Dataset
        """
        project_object = self._get_parent_project(project, workspace)
        return project_object.list_datasets(search, page_index, page_size)

    def delete_dataset(
        self, dataset: Reference, workspace: Optional[Reference] = None, project: Optional[Reference] = None
    ) -> None:
        """
        Deletes a dataset and all the versions that belong to it.

        :param dataset: The dataset name or id
        :param workspace: The workspace name or id the project belongs to
        :param project: The project name or id to get.

        :return: None
        """
        logging.getLogger("Project").propagate = False
        dataset_object = self.get_dataset(dataset, project, workspace)
        if dataset_object.project is None:
            raise SystemError("Project could not be found in dataset")
        logging.getLogger("Project").propagate = True
        dataset_object.project.delete_dataset(dataset_object.id)

    def delete_dataset_version(
        self,
        version: Reference,
        dataset: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> None:
        """
        Deletes a specific dataset in the specified project.

        :param dataset: dataset name or id
        :param version: The dataset version name or id
        :param workspace: The workspace name or id the project belongs to
        :param project: The project name or id to get.

        :return: None
        """
        logging.getLogger("Dataset").propagate = False
        dataset_version_object = self.get_dataset_version(version, dataset, workspace, project)
        if dataset_version_object.dataset is None:
            raise SystemError("Dataset could not be found in dataset version")
        logging.getLogger("Dataset").propagate = True
        dataset_version_object.dataset.delete_dataset_version(dataset_version_object.id)

    def update_dataset(
        self,
        dataset: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        connection: Optional[Reference] = None,
    ) -> Dataset:
        """
        Updates a dataset. The connection is associated with the resources, e.g gcs connection for a gcs resource.

        :param workspace: The workspace name or id the project belongs to
        :param project: The project name or id to get.
        :param dataset: dataset name or id
        :param name: The name of the dataset version e.g Changes 'Version 1' to 'name'
        :param description: The description of the dataset version
        :param connection: The connection name or id

        :return: the updated Dataset
        """
        logging.getLogger("Project").propagate = False
        dataset_object = self.get_dataset(dataset, project, workspace)
        if dataset_object.project is None:
            raise SystemError("Project not found for dataset")
        logging.getLogger("Project").propagate = True
        return dataset_object.project.update_dataset(dataset, name, description, connection)

    def update_dataset_version(
        self,
        version: Reference,
        dataset: Optional[Reference],
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        is_starred: Optional[bool] = None,
        properties: Optional[Dict] = None,
        resources: Optional[List[str]] = None,
        metadata: Optional[List[FileMetadata]] = None,
    ) -> DatasetVersion:
        """
        Updates a dataset version. The resource update is unique as if the hash is the same then the resource is
        not updated. If the hash is different then the resource will be updated.

        :param dataset: dataset name or id
        :param version: The dataset version name or id
        :param workspace: The workspace name or id the project belongs to
        :param project: The project name or id to get.
        :param name: The name of the dataset version e.g Changes 'Version 1' to 'name'
        :param description: The description of the dataset version
        :param is_starred: Whether the dataset version is starred
        :param properties: The properties of the dataset version
        :param resources: The resource/s uri/s of the dataset version
        :param metadata: The metadata uri/s of the dataset version

        :return: the updated Dataset Version
        """
        logging.getLogger("Project").propagate = False
        if isinstance(version, int):
            dataset_version_output = self._client.get_dataset_version(version)
            if dataset_version_output is None:
                raise InvalidReferenceError("version", version)
            parent_dataset = self.get_dataset(dataset_version_output.dataset_id)
        else:
            if dataset is None:
                raise MissingReferenceError("dataset version", "dataset")
            if project is None and self.project is not None:
                project = self.project.id
            if project is None:
                raise MissingReferenceError("job", "project")
            if workspace is None and isinstance(project, str) and self.workspace is not None:
                workspace = self.workspace.id
            parent_dataset = self.get_dataset(dataset, project, workspace)
        logging.getLogger("Project").propagate = True
        return parent_dataset.update_dataset_version(
            version, name, description, is_starred, properties, resources, metadata
        )

    def create_connection(
        self,
        name: str,
        type: ConnectionType,
        parameters: Optional[Dict[str, Any]] = None,
        description: str = "",
        workspace: Optional[Reference] = None,
    ):
        """
        Create a connection with the specified connection name, description and
        parameters. The parameters keys and values depends on each connection Type.
        Example of required parameters. {"key_value": data, "file_name": "file.json", "connection_info": data}


        :param name: name of the connection
        :param type: the type of connection
        :param parameters: the parameters for the connection
        :param description: the description of the connection
        :param workspace: The workspace name or id
        :return: a connection
        """

        if workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        if workspace is None:
            raise MissingReferenceError("connection", "workspace")
        parent_workspace = self.get_workspace(workspace)
        return parent_workspace.create_connection(name, type, parameters, description)

    def list_connections(
        self,
        workspace: Optional[Reference] = None,
        connection_type: Optional[str] = None,
        search: Optional[str] = None,
        page_index: int = 1,
        page_size: int = 20,
    ) -> List[Connection]:
        """
        List the connections in a workspace.

        List the users connections having access to the specified workspace.

        :param workspace: The workspace name or id
        :param search: The connection name to search

        :return: List of Connections
        """
        workspace_object = self.workspace if workspace is None else self.get_workspace(workspace)
        if workspace_object is None:
            raise MissingReferenceError("workspace")
        return workspace_object.list_connections(connection_type, search, page_index, page_size)

    def get_connection(self, connection: Reference, workspace: Optional[Reference] = None) -> Connection:
        """
        retrieve an existing connection

        :param connection: The connection name or id
        :param workspace: the workspace where the connection has been created
        :return: A Connection if it is known
        """
        workspace_object = self.workspace if workspace is None else self.get_workspace(workspace)
        if workspace_object is None:
            raise MissingReferenceError("workspace")
        return workspace_object.get_connection(connection)

    def delete_connection(self, connection: Reference, workspace: Optional[Reference] = None):
        """
        Deletes the connection specified by the user.

        :param workspace: The workspace name or id the project belongs to
        :param connection: The connection name or id to delete.

        :return: None
        """
        workspace_object = self.workspace if workspace is None else self.get_workspace(workspace)
        if workspace_object is None:
            raise MissingReferenceError("workspace")
        return workspace_object.delete_connection(connection)

    def update_connection(
        self,
        connection: Reference,
        workspace: Optional[Reference] = None,
        name: Optional[str] = None,
        parameters: Optional[Dict] = None,
        description: Optional[str] = None,
    ) -> Connection:
        workspace_object = self.workspace if workspace is None else self.get_workspace(workspace)
        if workspace_object is None:
            raise MissingReferenceError("workspace")
        result = workspace_object.update_connection(connection, name, description=description, parameters=parameters)
        return result

    def update_model_version(
        self,
        version: Reference,
        model: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        status: Optional[ModelVersionStatus] = ModelVersionStatus.EXPERIMENTATION,
        algorithm: Optional[str] = None,
        is_starred: Optional[bool] = False,
        metrics: Optional[Dict] = None,
        hyper_parameters: Optional[Dict] = None,
    ) -> ModelVersion:
        """
        Update a model version.

        ?Updates the specified model version with the new params passed to this function. ?

        :param model: The model name or id the version belongs to
        :param version: The model version name or id
        :param name: The name of the model version
        :param workspace: The workspace name or id the project belongs to
        :param project: The project name or id to get.
        :param description: The description of the model
        :param status: The status of the model e.g 'EXPERIMENTATION'/'STAGING'/'PRODUCTION'
        :param algorithm: The algorithm the model version uses
        :param is_starred: Whether the model is starred or not
        :param metrics: The model version metrics
        :param hyper_parameters: The model version hyper parameters

        :return: None
        """
        logging.getLogger("Project").propagate = False
        if isinstance(version, int):
            model_version_output = self._client.get_model_version(version)
            parent_model = self.get_model(model_version_output.model_id)
        elif model is not None:
            parent_model = self.get_model(model, project, workspace)
        else:
            raise MissingReferenceError("model version", "model")
        logging.getLogger("Project").propagate = True
        result: ModelVersion = parent_model.update_model_version(
            version, name, description, status, algorithm, is_starred, metrics, hyper_parameters
        )
        return result

    def list_model_versions(
        self,
        model: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        search: Optional[str] = None,
        page_index: int = Page.index,
        page_size: int = Page.size,
    ) -> List[ModelVersion]:
        """
        List the model versions of a model.

        Lists all the versions created out of the specified model.

        :param workspace: The workspace name or id the project belongs to
        :param project: The project name or id to get.
        :param model: The model name or id the versions belongs to

        :return: List of ModelVersions
        """
        model_object: Model = self.get_model(model, project, workspace)
        return model_object.list_model_versions(search, page_index, page_size)

    def list_model_versions_dataframe(
        self,
        model: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        search: Optional[str] = None,
        page_index: int = Page.index,
        page_size: int = Page.size,
    ) -> "pandas.DataFrame":  # type: ignore # noqa F821
        """
        Lists model versions in a pandas DataFrame and sorts by update date. Requires the pandas module to be installed,
        which can be done with ```pip install pandas``` or ```pip install vectice[pandas]```

        :param model: The id of the model
        :param workspace: The workspace name or id the project belongs to
        :param project: The project name or id to get.
        :param search: filter on name
        :param page_index: The page index. For example 1/20 would return page 1
        :param page_size: The number of versions on a page
        :return: pd.DataFrame
        """

        model_object: Model = self.get_model(model, project, workspace)
        return model_object.list_model_versions_dataframe(search, page_index, page_size)

    def get_model_version(
        self,
        version: Reference,
        model: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> ModelVersion:
        """
        Gets a specific version of a model.

        :param model: The model name or id the version belongs to
        :param version: The model version name or id
        :param workspace: The workspace name or id the project belongs to
        :param project: The project name or id to get.

        :return: A ModelVersion
        """

        if isinstance(version, int):
            model_version_output = self._client.get_model_version(version)
            if model_version_output is None:
                raise InvalidReferenceError("version", version)
            logging.getLogger("Project").propagate = False
            parent_model = self.get_model(model_version_output.model_id)
            logging.getLogger("Project").propagate = True
        elif model is not None:
            logging.getLogger("Project").propagate = False
            parent_model = self.get_model(model, project, workspace)
            logging.getLogger("Project").propagate = True
        else:
            raise MissingReferenceError("model version", "model")

        result: ModelVersion = parent_model.get_model_version(version)
        return result

    def delete_model_version(
        self,
        version: Reference,
        model: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> None:
        """
        Deletes the specified version of a model.

        :param workspace: The workspace name or id the project belongs to
        :param project: The project name or id to get.
        :param model: The model name or id the version belongs to
        :param version: The model version name or id

        :return: None
        """
        logging.getLogger("Project").propagate = False
        if isinstance(version, int):
            model_version_output = self._client.get_model_version(version)
            parent_model = self.get_model(model_version_output.model_id)
        else:
            if model is None:
                raise MissingReferenceError("dataset version", "dataset")
            if project is None and self.project is not None:
                project = self.project.id
            if project is None:
                raise MissingReferenceError("job", "project")
            if workspace is None and isinstance(project, str) and self.workspace is not None:
                workspace = self.workspace.id
            parent_model = self.get_model(model, project, workspace)
        logging.getLogger("Project").propagate = True
        parent_model.delete_model_version(version)

    def delete_model_version_metrics(
        self,
        version: Reference,
        metrics: List[str],
        model: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> None:
        """
        Delete a model version metrics.

        Delete a model version metrics.

        :param version: The version name or id of the model
        :param metrics: The metric keys to be deleted
        :param model: The model id or name
        :param project: The project name or id to get.
        :param workspace: The workspace name or id to get.

        :return:
        """
        model_version_object = self.get_model_version(version, model, project, workspace)
        model_version_object.model.delete_model_version_metrics(version, metrics)

    def delete_model_version_hyper_parameters(
        self,
        version: Reference,
        hyper_parameters: List[str],
        model: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> None:
        """
        Delete a model version hyper parameters.

        Delete a model version hyper parameters.

        :param version: The version name or id of the model
        :param hyper_parameters: The hyper parameter keys to be deleted
        :param model: The model id or name
        :param workspace: The workspace name or id
        :param project: The project name or id

        :return:
        """
        model_version_object = self.get_model_version(version, model, project, workspace)
        model_version_object.model.delete_model_version_properties(version, hyper_parameters)

    def create_model(
        self,
        name: str,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        description: Optional[str] = None,
        type: Optional[ModelType] = ModelType.OTHER,
    ) -> Model:
        """
        Create a model.

        Creates a new model in the specified project.

        :param name: The name of the model
        :param workspace: The workspace name or id
        :param project: The project name or id
        :param description: The description of the model
        :param type: The model type to use

        :return: A Model
        """
        if project is None and self.project is not None:
            project = self.project.id
            workspace = None
        elif isinstance(project, str) and workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        parent_project = self._get_parent_project(project, workspace)
        return parent_project.create_model(name, description, type)

    def update_model(
        self,
        model: Reference,
        workspace: Optional[Reference] = None,
        project: Optional[Reference] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        type: Optional[ModelType] = ModelType.OTHER,
    ) -> Model:
        """
        Update a model.

        :param model: The model name or id
        :param workspace: The workspace name or id
        :param project: The project name or id
        :param name: The name of the model
        :param description: The description of the model
        :param type: The model type

        :return: Model
        """
        logging.getLogger("Project").propagate = False
        if isinstance(model, int):
            model_object = self.get_model(model)
        elif isinstance(model, str):
            if project is None and self.project is not None:
                project = self.project.id
            if project is None:
                raise MissingReferenceError("model", "project")
            if workspace is None and isinstance(project, str) and self.workspace is not None:
                workspace = self.workspace.id
            model_object = self.get_model(model, project, workspace)
        else:
            raise InvalidReferenceError("model", model)
        logging.getLogger("Project").propagate = True
        result: Model = model_object.project.update_model(model, name, description, type)
        return result

    def list_models(
        self,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        search: Optional[str] = None,
        page_index: int = Page.index,
        page_size: int = Page.size,
    ) -> List[Model]:
        """
        List the models.

        Lists the models of the specified project, the search param helps find the models with the specified name.

        :param workspace: The workspace name or id
        :param project: The project name or id
        :param search: filter on name

        :return: List of Models
        """
        parent_project = self._get_parent_project(project, workspace)
        return parent_project.list_models(search, page_index, page_size)

    def delete_model(
        self,
        model: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> None:
        """
        Deletes the model with the specified reference.

        :param model: The model name or id
        :param workspace: The workspace name or id
        :param project: The project name or id

        :return: None
        """
        logging.getLogger("Project").propagate = False
        model_object = self.get_model(model, project, workspace)
        logging.getLogger("Project").propagate = True
        model_object.project.delete_model(model)

    def create_code_version(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        is_starred: Optional[bool] = None,
        uri: Optional[str] = None,
        git_version: Optional[GitVersion] = None,
        script_relative_path: str = ".",
        check_remote_repository: bool = True,
        attachments: Optional[Union[str, List[str]]] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> CodeVersion:
        """
        Create a code version.

        Creates a code version in the specified project using the git repository path of the code version.

        :param name: The name of the dataset version e.g Changes 'Version 1' to 'name'
        :param description: The description of the dataset version
        :param is_starred: Whether the dataset version is starred
        :param uri:
        :param git_version:
        :param script_relative_path: The path to the git repository
        :param check_remote_repository: indicate if the version should check the entry is existing on origin repository
        :param attachments: Resources to attach to the code version e.g AWS S3 or local files
        :param workspace: The workspace name or id
        :param project: The project name or id

        :return: A CodeVersion
        """
        logging.getLogger("Project").propagate = False
        if project is None and self.project is not None:
            project = self.project.id
        else:
            if project is None:
                raise MissingReferenceError("project")
            if isinstance(project, str) and workspace is None and self.workspace is not None:
                workspace = self.workspace.id
        if git_version is None:
            git_version = GitVersion.create(script_relative_path, check_remote_repository=check_remote_repository)
        project_object = self._get_parent_project(project, workspace)
        logging.getLogger("Project").propagate = True
        code_version = project_object.create_code_version(name, description, uri, is_starred, git_version)
        if attachments is not None:
            for attachment in attachments:
                code_version.add_attachments(attachment)
        return code_version

    def create_code_version_github_uri(
        self,
        uri: str,
        script_relative_path: str,
        login_or_token: Optional[str] = None,
        password: Optional[str] = None,
        jwt: Optional[str] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> CodeVersion:
        """
        Create a code version using an URI.

        :param uri: The uri of the repository with a specific branch if needed.
        :param script_relative_path:  The file that is executed
        :param login_or_token: A real login or a personal access token
        :param password: The password
        :param jwt: The OAuth2 access token
        :param workspace: The workspace name or id
        :param project: The project name or id

        :return: A CodeVersion
        """
        git_version = GitVersion.create_from_github_uri(uri, script_relative_path, login_or_token, password, jwt)
        if git_version is None:
            raise ValueError("Please check the GitHub uri and script relative path.")
        project_object = self._get_parent_project(project, workspace)
        return project_object.create_code_version(git_version=git_version)

    def create_code_version_gitlab_uri(
        self,
        uri: str,
        script_relative_path: str,
        private_token: Optional[str] = None,
        oauth_token: Optional[str] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> CodeVersion:
        """
        Create a code version using an URI.

        :param uri: The uri of the repository with a specific branch if needed.
        :param script_relative_path:  The file that is executed
        :param private_token: A real login or a personal access token
        :param oauth_token: The password
        :param workspace: The workspace name or id
        :param project: The project name or id

        :return: A CodeVersion
        """
        git_version = GitVersion.create_from_gitlab_uri(uri, script_relative_path, private_token, oauth_token)
        if git_version is None:
            raise ValueError("Please check the Gitlab uri and script relative path.")
        project_object = self._get_parent_project(project, workspace)
        return project_object.create_code_version(git_version=git_version)

    def create_code_version_bitbucket_uri(
        self,
        uri: str,
        script_relative_path: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> CodeVersion:
        """
        Create a code version using an URI.

        :param uri: The uri of the repository with a specific branch if needed.
        :param script_relative_path:  The file that is executed
        :param username: A real login or a personal access token
        :param password: The password
        :param workspace: The workspace name or id
        :param project: The project name or id

        :return: A CodeVersion
        """
        git_version = GitVersion.create_from_bitbucket_uri(uri, script_relative_path, username, password)
        if git_version is None:
            raise ValueError("Please check the BitBucket uri and script relative path.")
        parent_project = self._get_parent_project(project, workspace)
        return parent_project.create_code_version(git_version=git_version)

    def update_code_version(
        self,
        version: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        is_starred: Optional[bool] = None,
        repository_name: Optional[str] = None,
        branch_name: Optional[str] = None,
        commit_hash: Optional[str] = None,
        commit_comment: Optional[str] = None,
        commit_author_name: Optional[str] = None,
        commit_author_email: Optional[str] = None,
        is_dirty: Optional[bool] = None,
        uri: Optional[str] = None,
        entrypoint: Optional[str] = None,
        check_remote_repository: Optional[bool] = True,
    ) -> CodeVersion:
        """
        Updates the specified code version with the new given information in the params.
        If the code version id is given there is no need for specifying the workspace or project references.

        :param version: The code version name or id
        :param name: The name of the code version e.g Changes 'Version 1' to 'name'
        :param description: The description of the code version
        :param is_starred: Whether the code version is starred
        :param repository_name: The git repository name
        :param branch_name: The git branch name
        :param commit_hash: The git commit hash
        :param commit_comment: The git commit comment
        :param commit_author_email: The git commite author email
        :param commit_author_name: The git commit author name
        :param is_dirty: If the git commit is dirty
        :param uri: The git uri
        :param entrypoint: The entrypoint of the script
        :param check_remote_repository: Check the remote repository

        :return: The updated code version
        """
        logging.getLogger("Project").propagate = False
        git_version = None
        if entrypoint is None:
            if (
                repository_name is not None
                and branch_name is not None
                and commit_hash is not None
                and commit_comment is not None
                and commit_author_name is not None
                and commit_author_email is not None
                and is_dirty is not None
                and uri is not None
            ):
                git_version = GitVersion(
                    repository_name,
                    branch_name,
                    commit_hash,
                    commit_comment,
                    commit_author_name,
                    commit_author_email,
                    is_dirty,
                    uri,
                )
        elif entrypoint:
            git_version = GitVersion.create(
                entrypoint, check_remote_repository=check_remote_repository if check_remote_repository else False
            )
        if project is None and self.project is not None:
            project = self.project.id
        else:
            if project is None:
                raise MissingReferenceError("project")
            if isinstance(project, str) and workspace is None and self.workspace is not None:
                workspace = self.workspace.id
        project_object = self._get_parent_project(project, workspace)
        logging.getLogger("Project").propagate = True
        return project_object.update_code_version(version, name, description, is_starred, uri, git_version=git_version)

    def list_code_versions(
        self,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        search: Optional[str] = None,
        page_index: int = 1,
        page_size: int = 20,
    ) -> List[CodeVersion]:
        """
        Lists all the versions of a code artifact.

        :param workspace: The workspace name or id
        :param project: The project name or id
        :param search: Search code version term
        :param page_index: The page index
        :param page_size: The page size

        :return: List of CodeVersions
        """
        if project is None and self.project is not None:
            project = self.project.id
        else:
            if project is None:
                raise MissingReferenceError("project")
            if isinstance(project, str) and workspace is None and self.workspace is not None:
                workspace = self.workspace.id
        project_object = self._get_parent_project(project, workspace)
        return project_object.list_code_versions(search, page_index, page_size)

    def get_code_version(
        self,
        version: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> CodeVersion:
        """
        Gets the specified version of a code artifact. If the code version id is given
        there is no need for specifying the workspace or project references.

        :param version: The version name or id.
        :param workspace: The workspace name or id
        :param project: The project name or id

        :return: A CodeVersion
        """
        if isinstance(version, int):
            code_version_output = self._client.get_code_version(version)
            if code_version_output is None:
                raise InvalidReferenceError("code version", version)
            else:
                project_object = self._get_parent_project(code_version_output.code.project.id, workspace)
        elif isinstance(version, str) and project or self.project:
            if project is not None:
                project_object = self._get_parent_project(project, workspace)
            elif self.project is not None:
                project_object = self.project
            else:
                raise MissingReferenceError("code version", version)
        else:
            raise InvalidReferenceError("code version", version)
        return project_object.get_code_version(version)

    def delete_code_version(
        self,
        version: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> None:
        """
        Deletes the specified version of a code artifact. If the code version id is given
        there is no need for specifying the workspace or project references.


        :param workspace: The workspace name or id
        :param project: The project name or id
        :param version: The version name or id.

        :return: None
        """
        logging.getLogger("Project").propagate = False
        code_version = self.get_code_version(version, project, workspace)
        logging.getLogger("Project").propagate = False
        return code_version.project.delete_code_version(code_version.id)

    def create_stage(
        self,
        name: Optional[str] = None,
        status: Optional[StageStatus] = StageStatus.NotStarted,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> Stage:
        """
        Add a stage to the project. A documentation page is created with the title provided. The status can be 'NotStarted/InProgress/Completed'.

        :param name: The name of the stages documentation page.
        :param status: The status of the stage.
        :param project: The name or id of the project
        :param workspace: The name or id of the workspace

        :return: A Stage object
        """
        parent_project = self._get_parent_project(project, workspace)
        return parent_project.create_stage(name, status)

    def update_stage(
        self,
        stage: Reference,
        name: Optional[str] = None,
        status: Optional[StageStatus] = StageStatus.NotStarted,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> Stage:
        """
        Update a stage of the project. A documentation page is updated with the title and status provided. The status can be 'Draft/InProgress/Completed'.

        :param stage: The name or id of the stage.
        :param name: The name of the stages documentation page.
        :param status: The status of the stage.
        :param project: The name or id of the project
        :param workspace: The name or id of the workspace

        :return: A Stage object
        """
        if isinstance(stage, int):
            stage_object = self.get_stage(stage)
        elif isinstance(stage, str):
            if project is None and self.project is not None:
                project = self.project.id
            if project is None:
                raise MissingReferenceError("stage", "project")
            if workspace is None and isinstance(project, str) and self.workspace is not None:
                workspace = self.workspace.id
            stage_object = self.get_stage(stage, project, workspace)
        else:
            raise InvalidReferenceError("stage", stage)
        result: Stage = stage_object.project.update_stage(stage, name, status)
        return result

    def get_stage(
        self, stage: Reference, project: Optional[Reference] = None, workspace: Optional[Reference] = None
    ) -> Stage:
        """
        Get a stage of the project. A stage will be returned.

        :param stage: The name or id of the stage.
        :param project: The name or id of the project
        :param workspace: The name or id of the workspace

        :return: Stage
        """
        if isinstance(stage, int):
            stage_output = self._client.get_stage(stage)
            if stage_output is None:
                raise InvalidReferenceError("version", stage)
            project_object = self._get_parent_project(stage_output.project_id, None)
            if project_object:
                self._client._project = stage_output.project
        elif isinstance(stage, str):
            if project is None and self.project is not None:
                project = self.project.id
            if isinstance(project, str) and workspace is None and self.workspace is not None:
                workspace = self.workspace.id
            if project is None:
                raise MissingReferenceError("stage", "project")
            project_object = self._get_parent_project(project, workspace)
        else:
            raise MissingReferenceError("stage", "stage")
        return project_object.get_stage(stage)

    def list_stages(
        self,
        search: Optional[str] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        page_index: int = Page.index,
        page_size: int = Page.size,
    ) -> List[Stage]:
        """
        List all the stages of the project. A search term can be used.

        :param search: The search term to be used.
        :param project: The name or id of the project
        :param workspace: The name or id of the workspace
        :param page_index: The page index to be returned
        :param page_size: The page size to be returned

        :return: A list of stages
        """
        parent_project = self._get_parent_project(project, workspace)
        return parent_project.list_stages(search, page_index, page_size)

    def delete_stage(
        self, stage: Reference, project: Optional[Reference] = None, workspace: Optional[Reference] = None
    ) -> None:
        """
        Delete a stage of the project. A documentation page related to the stage will be deleted aswell.

        :param stage: The name or id of the stage.
        :param project: The name or id of the project
        :param workspace: The name or id of the workspace


        :return: None
        """
        parent_project = self._get_parent_project(project, workspace)
        return parent_project.delete_stage(stage)

    def document_run(self, run_id: Optional[int] = None, name: Optional[str] = None, **kwargs) -> None:
        """
        Captures all the assets of a run in a stage and documentation.

        :param run_id: The id of the run.
        :param name: The name of the stage.

        :return: None
        """
        logger.info("Run will be documented and captured in a Stage.")
        if run_id is not None:
            """
            Captures an existing run. That has been completed.
            """
            run = self.get_run(run_id)
            if run and run.status != RunStatus.COMPLETED:
                logging.warning(
                    "The run is not in a completed state and the auto documentation might not function as expected."
                )
            if name is not None:
                try:
                    stage = self.get_stage(name)
                except Exception:
                    logger.debug(f"Stage {name} for auto documentation not found.")
                    stage = self.create_stage(name, status=StageStatus.InProgress)
            else:
                stage = self.create_stage(f"Auto Document {run.name}", status=StageStatus.InProgress)
            run._capture_run_info(stage)
        elif run_id is None and len(self._active_runs) >= 1 or run_id is None and kwargs.get("_experiment_run"):
            """
            Sets up to capture the run upon completion. Experiment.complete() / vectice.end_run()
            """
            self._auto_document_run = True
            self._document_stage = name
        else:
            raise MissingReferenceError("No run provided and no active run.")



from __future__ import annotations

from typing import Optional, List
from typing import TYPE_CHECKING
from urllib.parse import urlencode

from ._auth import Auth
from ._http import HttpError
from .http_error_handlers import HttpErrorHandler
from .json import BlockInput, BlockOutput, Page
from .stage import StageApi
from .reference import InvalidReferenceError

if TYPE_CHECKING:
    from vectice import Reference


class BlockApi:
    def __init__(self, auth: Auth):
        self._auth = auth
        self._httpErrorhandler = HttpErrorHandler()

    def create_block(
        self, stage: Reference, data: BlockInput, project: Reference, workspace: Optional[Reference] = None
    ) -> BlockOutput:
        if isinstance(stage, int):
            url = f"/metadata/stages/{stage}/file/blocks"
        elif isinstance(stage, str):
            stage_object = StageApi(self._auth).get_stage(stage, project, workspace)
            url = f"/metadata/stages/{stage_object.id}/file/blocks"
        else:
            raise InvalidReferenceError("stage", stage)
        try:
            response = self._auth._post(url, data)
            return BlockOutput(**response)
        except HttpError as e:
            raise self._httpErrorhandler.handlePostHttpError(e, "block")

    def update_block(
        self,
        position: int,
        stage: Reference,
        data: BlockInput,
        project: Reference,
        workspace: Optional[Reference] = None,
    ) -> BlockOutput:
        if isinstance(stage, int):
            url = f"/metadata/stages/{stage}/file/blocks/{position}"
        elif isinstance(stage, str):
            stage_object = StageApi(self._auth).get_stage(stage, project, workspace)
            url = f"/metadata/stages/{stage_object.id}/file/blocks/{position}"
        else:
            raise InvalidReferenceError("stage", stage)
        try:
            response = self._auth._put(url, data)
            return BlockOutput(**response)
        except HttpError as e:
            raise self._httpErrorhandler.handlePutHttpError(e, "stage", stage)

    def list_blocks(
        self,
        stage: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        search: Optional[str] = None,
        page_index=Page.index,
        page_size=Page.size,
    ) -> List[BlockOutput]:
        if isinstance(stage, int):
            base_url = f"/metadata/stages/{stage}/file"
        elif isinstance(stage, str):
            stage_output = StageApi(self._auth).get_stage(stage, project, workspace)
            base_url = f"/metadata/stages/{stage_output.id}/file"
        else:
            raise InvalidReferenceError("stage", stage)
        try:
            queries = {"index": page_index, "size": page_size}
            if search:
                queries["search"] = search
            response = self._auth._get(f"{base_url}?{urlencode(queries)}")
            result = [BlockOutput(item) for item in response["blocks"]]
            return result
        except HttpError as e:
            raise self._httpErrorhandler.handleGetHttpError(e, "stage", stage)

    def delete_block(
        self,
        position: int,
        stage: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> None:
        try:
            if isinstance(stage, int):
                self._auth._delete(f"/metadata/stages/{stage}/file/blocks/{position}")
            else:
                stage_object = StageApi(self._auth).get_stage(stage, project, workspace)
                self._auth._delete(f"/metadata/stages/{stage_object.id}/file/blocks/{position}")
        except HttpError as e:
            raise self._httpErrorhandler.handleGetHttpError(e, "stage", stage)



from typing import Optional
from .project import ProjectOutput
from enum import Enum


class BlockType(Enum):
    """
    Indicates the Documentation Page Origin.
    """

    EDITABLE = "EDITABLE"
    """
    """
    VECTICE_RESOURCE = "VECTICE_RESOURCE"
    """
    """
    CLOUD_FILE = "CLOUD_FILE"
    """
    """
    VERSIONED_FILE = "VERSIONED_FILE"
    """
    """
    OTHER = "OTHER"
    """
    """


class BlockVariant(Enum):
    """
    Indicates the Documentation Page Status.
    """

    TEXT = "TEXT"
    """
    """
    MODEL = "MODEL"
    """
    """
    MODEL_VERSION = "MODEL_VERSION"
    """
    """
    DATASET = "DATASET"
    """
    """
    DATASET_VERSION = "DATASET_VERSION"
    """
    """
    RUN = "RUN"
    """
    """
    IMAGE = "IMAGE"
    """
    """
    NOTEBOOK = "NOTEBOOK"
    """
    """
    FILE = "FILE"
    """
    """
    FORMULA = "FORMULA"
    """
    """
    CODE = "CODE"
    """
    """
    DIVIDER = "DIVIDER"
    """
    """


class BlockState(Enum):
    """
    Indicates the Block State.
    """

    COLLAPSE = "COLLAPSE"
    """
    """
    SMALL = "SMALL"
    """
    """
    MEDIUM = "MEDIUM"
    """
    """
    FULL = "FULL"
    """
    """


class BlockInput(dict):
    @property
    def type(self) -> str:
        return str(self["type"])

    @property
    def variant(self) -> str:
        return str(self["variant"])

    @property
    def content(self) -> Optional[str]:
        return str(self["content"])

    @property
    def name(self) -> Optional[str]:
        return str(self["name"])

    @property
    def position(self) -> Optional[int]:
        return int(self["position"])

    @property
    def searchable_content(self) -> Optional[str]:
        return str(self["searchableContent"])

    @property
    def state(self) -> Optional[str]:
        return str(self["state"])

    @property
    def resource_id(self) -> Optional[int]:
        return int(self["resourceId"])

    @property
    def entity_file_id(self) -> Optional[int]:
        return int(self["entityFileId"])

    @property
    def file_size(self) -> Optional[int]:
        return int(self["fileSize"])

    @property
    def file_path(self) -> Optional[str]:
        return str(self["filePath"])

    @property
    def url(self) -> Optional[str]:
        return str(self["url"])

    @property
    def config(self) -> Optional[str]:
        return str(self["config"])


class BlockOutput(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if "project" in self:
            self._project: ProjectOutput = ProjectOutput(**self["project"])
        else:
            self._project = None

    def items(self):
        result = []
        for key in self:
            if self[key] is not None:
                result.append((key, self[key]))
        return result

    @property
    def id(self) -> int:
        return int(self["id"])

    @property
    def type(self) -> str:
        return str(self["type"])

    @property
    def variant(self) -> str:
        return str(self["variant"])

    @property
    def content(self) -> Optional[str]:
        return str(self["content"])

    @property
    def name(self) -> Optional[str]:
        return str(self["name"])

    @property
    def position(self) -> Optional[int]:
        return int(self["position"])

    @property
    def searchable_content(self) -> Optional[str]:
        return str(self["searchableContent"])

    @property
    def state(self) -> Optional[str]:
        if self.get("state", None):
            return str(self["state"])
        else:
            return None

    @property
    def resource_id(self) -> Optional[int]:
        if self.get("resourceId", None):
            return int(self["resourceId"])
        else:
            return None

    @property
    def entity_file_id(self) -> Optional[int]:
        if self.get("entityFileId", None):
            return int(self["entityFileId"])
        else:
            return None

    @property
    def file_size(self) -> Optional[int]:
        if self.get("fileSize", None):
            return int(self["fileSize"])
        else:
            return None

    @property
    def file_path(self) -> Optional[str]:
        if self.get("filePath", None):
            return str(self["filePath"])
        else:
            return None

    @property
    def url(self) -> Optional[str]:
        return str(self["url"])

    @property
    def config(self) -> Optional[str]:
        return str(self["config"])



from __future__ import annotations

from datetime import datetime, timezone
from typing import Optional, Dict, Union, List


def create_properties(properties: Dict[str, Union[int, float, str]]) -> List[Property]:
    if len(set(properties)) < len(properties):
        raise ValueError("You can not use the same key value pair more than once.")
    props: List[Property] = []
    for key, value in properties.items():
        _check_empty_property(key, value)
        props.append(Property(key, value))
    return props


def _check_empty_property(key: Optional[str], value: Union[str, int, float, None]):
    if key is None or key.strip() == "" or value is None or (isinstance(value, str) and value.strip() == ""):
        raise ValueError("Property keys and values can't be empty.")


class Property:
    def __init__(self, key, value, name=None):
        self._key: str = key
        """"""
        self._value: str = value
        """"""
        self._timestamp: Union[datetime, str] = datetime.now(timezone.utc).isoformat()
        """"""
        self._name: Optional[str] = None
        """"""

    def __repr__(self):
        return f"Property(key={self.key}, value={self.value}, timestamp={self.timestamp}, name={self.name})"

    @property
    def key(self):
        return self._key

    @property
    def value(self):
        return self._value

    @property
    def timestamp(self):
        return self._timestamp

    @property
    def name(self):
        return self._name



from __future__ import annotations

import urllib
from typing import Optional, List
from typing import TYPE_CHECKING
from urllib.parse import urlencode

from ._auth import Auth
from ._http import HttpError
from .http_error_handlers import HttpErrorHandler
from .json import StageInput, StageOutput, Page
from .project import ProjectApi
from .reference import InvalidReferenceError, MissingReferenceError

if TYPE_CHECKING:
    from vectice import Reference


class StageApi:
    def __init__(self, auth: Auth):
        self._auth = auth
        self._httpErrorhandler = HttpErrorHandler()

    def create_stage(self, data: StageInput, project: Reference, workspace: Optional[Reference] = None) -> StageOutput:
        if isinstance(project, int):
            url = f"/metadata/project/{project}/stages"
        elif isinstance(project, str):
            parent_project = ProjectApi(self._auth).get_project(project, workspace)
            url = f"/metadata/project/{parent_project.id}/stages"
        else:
            raise InvalidReferenceError("project", project)
        try:
            response = self._auth._post(url, data)
            return StageOutput(**response)
        except HttpError as e:
            raise self._httpErrorhandler.handlePostHttpError(e, "stage")

    def update_stage(
        self,
        data: StageInput,
        stage: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> StageOutput:
        if data.name is None:
            raise ValueError('"name" must be provided in dataset.')
        if isinstance(stage, int):
            url = f"/metadata/stages/{stage}"
        elif isinstance(stage, str):
            stage_object = self.get_stage(stage, project, workspace)
            url = f"/metadata/stages/{stage_object.id}"
        else:
            raise InvalidReferenceError("stage", stage)
        try:
            response = self._auth._put(url, data)
            return StageOutput(**response)
        except HttpError as e:
            raise self._httpErrorhandler.handlePutHttpError(e, "stage", stage)

    def get_stage(
        self, stage: Reference, project: Optional[Reference] = None, workspace: Optional[Reference] = None
    ) -> StageOutput:
        if not isinstance(stage, int) and not isinstance(stage, str):
            raise InvalidReferenceError("stage", stage)
        if isinstance(stage, int):
            url = f"/metadata/stages/{stage}"
        else:
            if project is None:
                raise MissingReferenceError("project")
            parent_project = ProjectApi(self._auth).get_project(project, workspace)
            url = f"/metadata/project/{parent_project.id}/stages/{urllib.parse.quote(stage)}"
        try:
            response = self._auth._get(url)
            result = StageOutput(response)
            return result
        except HttpError as e:
            raise self._httpErrorhandler.handleGetHttpError(e, "stage", stage)
        except IndexError:
            raise ValueError("The stage is invalid. Please check the entered value.")

    def list_stages(
        self,
        project: Reference,
        workspace: Optional[Reference] = None,
        search: Optional[str] = None,
        page_index=Page.index,
        page_size=Page.size,
    ) -> List[StageOutput]:
        if isinstance(project, int):
            base_url = f"/metadata/project/{project}/stages"
        elif isinstance(project, str):
            parent_project = ProjectApi(self._auth).get_project(project, workspace)
            base_url = f"/metadata/project/{parent_project.id}/stages"
        else:
            raise InvalidReferenceError("project", project)
        try:
            queries = {"index": page_index, "size": page_size}
            if search:
                queries["search"] = search
            response = self._auth._get(f"{base_url}?{urlencode(queries)}")
            result = [StageOutput(item) for item in response]
            return result
        except HttpError as e:
            raise self._httpErrorhandler.handleGetHttpError(e, "project", project)

    def delete_stage(
        self, stage: Reference, project: Optional[Reference] = None, workspace: Optional[Reference] = None
    ) -> None:
        try:
            if isinstance(stage, str):
                stage_object = self.get_stage(stage, project, workspace)
                self._auth._delete(f"/metadata/stages/{stage_object.id}")
            else:
                self._auth._delete(f"/metadata/stages/{stage}")
        except HttpError as e:
            raise self._httpErrorhandler.handleGetHttpError(e, "stage", stage)



from __future__ import annotations

from typing import Optional, TYPE_CHECKING, List
from urllib.parse import urlencode

from ._auth import Auth
from ._http import HttpError
from .code import CodeApi
from .http_error_handlers import HttpErrorHandler
from .json import Page, PagedResponse, CodeVersionInput, CodeVersionOutput
from .project import ProjectApi
from .reference import InvalidReferenceError, MissingReferenceError

if TYPE_CHECKING:
    from vectice import Reference


class CodeVersionApi:
    def __init__(self, auth: Auth):
        self._auth = auth
        self._httpErrorhandler = HttpErrorHandler()

    def create_code_version(
        self,
        code_data: CodeVersionInput,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> CodeVersionOutput:
        if project is None:
            raise InvalidReferenceError("project", project)
        project_output = ProjectApi(self._auth).get_project(project, workspace)
        url = f"/metadata/project/{project_output.id}/codeversion"
        try:
            response = self._auth._post(url, code_data)
            return CodeVersionOutput(**response)
        except HttpError as e:
            raise self._httpErrorhandler.handlePostHttpError(e, "code")

    def get_code_version(
        self,
        version: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> CodeVersionOutput:
        if not isinstance(version, int) and not isinstance(version, str):
            raise ValueError("The code version reference is invalid. Please check the entered value.")
        if isinstance(version, int):
            try:
                url = f"/metadata/codeversion/{version}"
                response = self._auth._get(url)
                return CodeVersionOutput(**response)
            except HttpError as e:
                raise self._httpErrorhandler.handleGetHttpError(e, "code version", version)
            except IndexError:
                raise ValueError("The code version is invalid. Please check the entered value.")
        elif isinstance(version, str):
            if project is None:
                raise MissingReferenceError("project", project)
            code_version_list = self.list_code_versions(project, workspace, version)
            if len(code_version_list) == 1:
                return code_version_list[0]
            else:
                raise ValueError("The code version is invalid. Please check the entered value.")

    def list_code_versions(
        self,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        search: Optional[str] = None,
        page_index=Page.index,
        page_size=Page.size,
    ) -> List[CodeVersionOutput]:
        if project is None:
            raise InvalidReferenceError("project", project)
        project_output = ProjectApi(self._auth).get_project(project, workspace)
        if search is not None:
            code_list = CodeApi(self._auth).list_codes(project, workspace).list
            for code in code_list:
                url = f"/metadata/project/{project_output.id}/code/{code.id}/version"
                queries = {"index": page_index, "size": page_size}
                if search:
                    queries["search"] = search
                response = self._auth._get(f"{url}?{urlencode(queries)}")
                result = PagedResponse(
                    item_cls=CodeVersionOutput, total=response["total"], page=response["page"], items=response["items"]
                )
                if len(result.list) >= 1:
                    return result.list
            else:
                raise ValueError("The code version was not found.")
        elif search is None:
            results = []
            code_list = CodeApi(self._auth).list_codes(project, workspace).list
            for code in code_list:
                url = f"/metadata/project/{project_output.id}/code/{code.id}/version"
                queries = {"index": page_index, "size": page_size}
                if search:
                    queries["search"] = search
                response = self._auth._get(f"{url}?{urlencode(queries)}")
                result = PagedResponse(
                    item_cls=CodeVersionOutput, total=response["total"], page=response["page"], items=response["items"]
                )
                if len(result.list) >= 1:
                    results += result.list
            return results

    def update_code_version(
        self,
        code_data: CodeVersionInput,
        version: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> CodeVersionOutput:
        if project is None:
            raise InvalidReferenceError("project", project)
        code_version_object = self.get_code_version(version, project, workspace)
        url = f"/metadata/codeversion/{code_version_object.id}"
        try:
            response = self._auth._put(url, code_data)
            return CodeVersionOutput(**response)
        except HttpError as e:
            raise self._httpErrorhandler.handlePutHttpError(e, "version", code_version_object.id)

    def delete_code_version(
        self,
        version: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> None:
        try:
            code_version_object = self.get_code_version(version, project, workspace)
            if code_version_object is None:
                raise ValueError("The code version was not found.")
            self._auth._delete(f"/metadata/codeversion/{code_version_object.id}")
        except HttpError as e:
            raise self._httpErrorhandler.handleGetHttpError(e, "code_version", version)



from dataclasses import dataclass
from datetime import datetime
from typing import Optional

from .artifact_reference import RulesDatasetVersionInput, RulesModelVersionInput, RulesCodeVersionInput
from .artifact_type import ArtifactType, JobArtifactType
from .dataset_version import DatasetVersionInput
from .model_version import ModelVersionInput


@dataclass
class ArtifactInput:
    artifactType: ArtifactType
    jobArtifactType: JobArtifactType
    dataset: Optional[RulesDatasetVersionInput] = None
    dataSetVersion: Optional[DatasetVersionInput] = None
    dataSetVersionId: Optional[int] = None
    model: Optional[RulesModelVersionInput] = None
    modelVersion: Optional[ModelVersionInput] = None
    modelVersionId: Optional[int] = None
    codeVersionId: Optional[int] = None
    description: Optional[str] = None
    metadataSource: Optional[str] = None
    authorId: Optional[int] = None
    jobRunId: Optional[int] = None
    deletedDate: Optional[datetime] = None
    createdDate: Optional[datetime] = None
    updatedDate: Optional[datetime] = None
    version: Optional[str] = None
    id: Optional[int] = None
    code: Optional[RulesCodeVersionInput] = None


class ArtifactOutput(dict):
    def items(self):
        result = []
        for key in self:
            if self[key] is not None:
                result.append((key, self[key]))
        return result

    @property
    def id(self) -> int:
        return int(self["id"])

    @property
    def artifact_id(self) -> int:
        if self.artifact_type == ArtifactType.DATASET:
            return int(self["dataSetVersionId"])
        elif self.artifact_type == ArtifactType.MODEL:
            return int(self["modelVersionId"])
        else:
            return int(self["codeVersionId"])

    @property
    def job_artifact_type(self) -> JobArtifactType:
        return JobArtifactType[self["jobArtifactType"]]

    @property
    def artifact_type(self) -> ArtifactType:
        return ArtifactType[self["artifactType"]]



from datetime import datetime
from typing import Optional, cast

from vectice.api._utils import read_nodejs_date
from .dataset import DatasetOutput


class DatasetVersionInput(dict):
    @property
    def dataset_id(self) -> int:
        return int(self["datasetId"])

    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def description(self) -> str:
        return str(self["description"])

    @property
    def uri(self) -> str:
        return str(self["uri"])

    @property
    def resources(self) -> str:
        return str(self["resources"])

    @property
    def files_metadata(self) -> str:
        return str(self["filesMetadata"])

    @property
    def is_starred(self) -> str:
        return str(self["isStarred"])

    @property
    def properties(self) -> str:
        return str(self["properties"])


class DatasetVersionOutput(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if "dataSet" in self:
            self._dataset: DatasetOutput = DatasetOutput(**self["dataSet"])
        else:
            self._dataset = None

    def items(self):
        result = []
        for key in self:
            if self[key] is not None:
                result.append((key, self[key]))
        return result

    @property
    def created_date(self) -> Optional[datetime]:
        return read_nodejs_date(str(self["createdDate"]))

    @property
    def updated_date(self) -> Optional[datetime]:
        return read_nodejs_date(str(self["updatedDate"]))

    @property
    def id(self) -> int:
        return int(self["id"])

    @property
    def version(self) -> int:
        return int(self["version"])

    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def description(self) -> Optional[str]:
        if self.get("description", None):
            return str(self["description"])
        else:
            return None

    @property
    def uri(self) -> str:
        return str(self["uri"])

    @property
    def author_id(self) -> int:
        return int(self["authorId"])

    @property
    def deleted_date(self) -> Optional[datetime]:
        return read_nodejs_date(str(self["deletedDate"]))

    @property
    def version_type(self) -> int:
        return int(self["versionType"])

    @property
    def version_number(self) -> int:
        return int(self["versionNumber"])

    @property
    def dataset_id(self) -> int:
        return int(self["dataSetId"])

    @property
    def version_folder_id(self) -> int:
        return int(self["versionFolderId"])

    @property
    def origin_id(self) -> int:
        return int(self["originId"])

    @property
    def is_starred(self) -> bool:
        return bool(self["isStarred"])

    @property
    def dataset(self) -> DatasetOutput:
        return self._dataset

    @property
    def files_metadata(self) -> dict:
        return cast(dict, self["filesMetadata"])

    @property
    def reused_version(self) -> bool:
        return "reusedVersion" in self and bool(self["reusedVersion"])



from __future__ import annotations

from typing import Optional

from ._auth import Auth
from ._http import HttpError
from .http_error_handlers import HttpErrorHandler
from .json import RunStatus, RunOutput, StartRunInput, StopRunOutput, StopRunInput
from .json.rule import FillRunInput, FillRunOutput
from .project import ProjectApi
from .reference import Reference
from .run import RunApi


class RuleApi:
    def __init__(self, auth: Auth):
        self._auth = auth
        self._httpErrorHandler = HttpErrorHandler()

    def start_run(
        self,
        data: StartRunInput,
        project: Reference,
        workspace: Optional[Reference] = None,
    ) -> RunOutput:
        data.run.status = RunStatus.STARTED
        try:
            parent_project = ProjectApi(self._auth).get_project(project, workspace)
            url = f"/metadata/project/{parent_project.id}/rules/start-run"
            try:
                response = self._auth._post(url, data)
                return RunOutput(**response)
            except HttpError as e:
                raise self._httpErrorHandler.handlePostHttpError(e, "run", "start")
        except HttpError as e:
            raise self._httpErrorHandler.handleGetHttpError(e, "project", project)

    def stop_run(self, data: StopRunInput) -> StopRunOutput:
        if data.jobRun.id is None:
            raise ValueError('"id" must be provided in run.')
        try:
            run_output = RunApi(self._auth).get_run(data.jobRun.id)
            url = f"/metadata/project/{run_output.job.project.id}/rules/stop-run"
            response = self._auth._post(url, data)
            return StopRunOutput(**response)
        except HttpError as e:
            raise self._httpErrorHandler.handlePostHttpError(e, "run", "stop")

    def fill_run(self, data: FillRunInput):
        if data.jobRunId is None:
            raise ValueError('"jobRunId" must be provided in input.')
        try:
            run_output = RunApi(self._auth).get_run(data.jobRunId)
            url = f"/metadata/project/{run_output.job.project.id}/rules/fill-run"
            response = self._auth._post(url, data)
            return FillRunOutput(**response)
        except HttpError as e:
            raise self._httpErrorHandler.handlePostHttpError(e, "run", "fill")



import logging
import os
from typing import BinaryIO, Union, List

from vectice.api import Client
from vectice.api.json.attachment import AttachmentOutput


class AttachmentContainer:
    def __init__(self, name: str, id: int, client: Client, container_type: str):
        self._client = client
        self._id = id
        self._name = name
        self._container_type = container_type

    def update_attachments(self, file_path: Union[str, List[str]]):
        """
        add an attachment to the entity

        :param file_path: the path to the attachment
        """
        attachments = []
        if isinstance(file_path, list):

            for file in file_path:
                try:
                    curr_file = ("file", (file, open(file, "rb")))
                    attachments.append(curr_file)
                except Exception:
                    logging.warning(f"Did not read {file}.")
        else:
            attachments = [("file", (file_path, open(file_path, "rb")))]
        self._client.update_attachments(self._container_type.lower(), attachments, self._id)

    def list_attachments(self) -> List[str]:
        """
        List attachments of the entity

        """
        attachments = self._client.list_attachments(self._container_type.lower(), self._id)
        files = []
        for attachment in attachments.list:
            files.append(attachment.fileName)
        return files

    def delete_attachments(self, file_path: Union[List[str], str]):
        """
        remove an attachment from the entity

        :param file_path: the path to the attachment
        """
        attachments: List[str] = []
        if isinstance(file_path, str):
            attachments.append(file_path)
        else:
            attachments = file_path

        try:
            attachment_list = {
                attach.fileName: attach.fileId
                for attach in self._client.list_attachments(self._container_type.lower(), self._id).list
            }
        except Exception as e:
            raise ValueError(f"{self._container_type} attachment failed to retrieve. Due to {e}")
        for file in attachments:
            file_id = attachment_list.get(file)
            if file_id:
                self._client.delete_attachment(self._container_type.lower(), self._id, file_id)
            else:
                raise ValueError(f"The file path for {file} is not valid. The file does not exist.")

    def add_attachments(self, file_path: Union[str, List[str]]):
        """
        add an attachment to the entity

        :param file_path: the path to the attachment
        """
        attachments = []
        attached_files = self.list_attachments()
        if isinstance(file_path, list):
            for file in file_path:
                try:
                    if not os.path.exists(file):
                        raise ValueError(f"the file path {file} is not valid. the file does not exist")
                    curr_file = ("file", (file, open(file, "rb")))
                    attachments.append(curr_file)
                    file_name = file.split("/")[-1]
                    if file_name in attached_files:
                        raise RuntimeError(f"{file_path} is already attached to '{self._name}'")
                except Exception:
                    logging.warning(f"Did not read {file}.")
        else:
            if not os.path.exists(file_path):
                raise ValueError(f"The file path {file_path} is not valid. The file does not exist.")
            attachments = [("file", (file_path, open(file_path, "rb")))]
            file_name = file_path.split("/")[-1]
            if file_name in attached_files:
                raise RuntimeError(f"{file_path} is already attached to '{self._name}'")
        self._client.create_attachments(self._container_type.lower(), attachments, self._id)

    def get_attachment(self, file_path: str) -> BinaryIO:
        """
        Get an attachment content from the entity

        :param file_path: the path to the attachment

        :return: BinaryIO
        """
        try:
            attachment_list = {
                attach.fileName: attach.fileId
                for attach in self._client.list_attachments(self._container_type.lower(), self._id).list
            }
        except Exception as e:
            raise ValueError(f"list of attachment for {self._container_type} failed . Due to {e}")
        file_id = attachment_list.get(file_path)
        if file_id:
            try:
                return self._client.get_attachment(self._container_type.lower(), file_id, self._id)
            except Exception as e:
                raise ValueError(
                    f"{self._container_type} attachment failed to retrieve attachment named '{file_path}' due to:\n {e}"
                )
        else:
            raise ValueError(
                f"{self._container_type} attachment failed to retrieve attachment named '{file_path}'. Please check the filename."
            )

    def get_attachment_as_file(self, file_path: str, saved_path: str):
        """
        Get an attachment content from the entity and store it in a file

        :param file_path: the path to the file
        :param saved_path: the path to the file

        :return: None
        """
        attachment = self.get_attachment(file_path)
        with open(saved_path, "wb") as content:
            while True:
                chunk = attachment.read(128)
                if not chunk:
                    break
                content.write(chunk)

    def _list_attachments(self) -> List[AttachmentOutput]:
        """
        List attachments of the entity

        """
        return self._client.list_attachments(self._container_type.lower(), self._id).list



from typing import Union, Any, Optional

from vectice.api._http import HttpError

Reference = Union[str, int]


class InvalidReferenceError(ValueError):
    def __init__(self, reference_type: str, value: Any) -> None:
        super().__init__(
            f"The {reference_type} reference is invalid. Please check the provided value. "
            + "it should be a string or a number. "
            + f"Provided value is {value} ({type(value)})"
        )


class BadReferenceError(ValueError):
    def __init__(self, reference_type: str, value: Reference, error: HttpError) -> None:
        if isinstance(value, str):
            super().__init__(
                # f"{error.method.upper()} {error.path}: HTTP CODE {error.code}\n"
                f"The reference does not reference any known {reference_type} . Please check the provided value. "
                + f"{reference_type} named '{value}' is unknown"
            )
        else:
            super().__init__(
                f"The reference does not reference any known {reference_type} . Please check the provided value. "
                + f"{reference_type} with identifier <{value}> is unknown"
            )


class MissingReferenceError(ValueError):
    def __init__(self, reference_type: str, parent_reference_type: Optional[str] = None) -> None:
        if parent_reference_type is not None:
            super().__init__(f"the {parent_reference_type} reference is required if the {reference_type} name is given")
        else:
            super().__init__(f"the {reference_type} reference is required")



from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Optional, List

from .artifact_version import ArtifactVersion
from .artifact import ArtifactInput
from .artifact_reference import (
    ArtifactReferenceInput,
    RulesModelVersionInput,
    RulesDatasetVersionInput,
    RulesCodeVersionInput,
)
from .job import JobInput, JobOutput
from .._utils import read_nodejs_date


class RunStatus(Enum):

    """
    Status of a Run.
    """

    SCHEDULED = "SCHEDULED"
    """"""
    STARTED = "STARTED"
    """"""
    COMPLETED = "COMPLETED"
    """"""
    FAILED = "FAILED"
    """"""
    ABORTED = "ABORTED"
    """"""


@dataclass
class RunInput:
    name: Optional[str]
    job: Optional[JobInput] = None
    systemName: Optional[str] = ""
    startDate: Optional[datetime] = None
    endDate: Optional[datetime] = None
    status: RunStatus = RunStatus.SCHEDULED
    inputs: Optional[List[ArtifactInput]] = None
    properties: Optional[dict] = None
    description: Optional[str] = None


class RunOutput(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if "job" in self:
            self._job: JobOutput = JobOutput(**self["job"])
        if "jobArtifacts" in self:
            self._inputs = [
                ArtifactReferenceInput(
                    item["artifactType"],
                    item.get("description", None),
                    dataset=RulesDatasetVersionInput(
                        parentId=item["dataSetVersion"]["dataSetId"],
                        version=ArtifactVersion(
                            version_number=item["dataSetVersion"]["versionNumber"],
                            version_id=item["dataSetVersion"]["id"],
                            version_name=item["dataSetVersion"]["name"],
                        ),
                    )
                    if item["artifactType"] == "DATASET"
                    else None,
                    model=RulesModelVersionInput(
                        parentId=item["modelVersion"]["modelId"],
                        version=ArtifactVersion(
                            version_number=item["modelVersion"]["versionNumber"],
                            version_id=item["modelVersion"]["id"],
                            version_name=item["modelVersion"]["name"],
                        ),
                    )
                    if item["artifactType"] == "MODEL"
                    else None,
                    code=RulesCodeVersionInput(
                        parentId=item["codeVersion"]["codeId"],
                        version=ArtifactVersion(
                            version_number=item["codeVersion"]["versionNumber"],
                            version_id=item["codeVersion"]["id"],
                            version_name=item["codeVersion"]["name"],
                        ),
                    )
                    if item["artifactType"] == "CODE"
                    else None,
                )
                for item in self["jobArtifacts"]
                if item["jobArtifactType"] == "INPUT"
            ]

    def items(self):
        result = []
        for key in self:
            if self[key] is not None:
                result.append((key, self[key]))
        return result

    @property
    def id(self) -> int:
        return int(self["id"])

    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def status(self) -> RunStatus:
        return RunStatus[self["status"]]

    @property
    def job_id(self) -> int:
        return int(self["jobId"])

    @property
    def start_date(self) -> Optional[datetime]:
        if "startDate" in self and self["startDate"] is not None:
            return read_nodejs_date(str(self["startDate"]))
        else:
            return None

    @property
    def end_date(self) -> Optional[datetime]:
        if "endDate" in self and self["endDate"] is not None:
            return read_nodejs_date(str(self["endDate"]))
        else:
            return None

    @property
    def created_date(self) -> Optional[datetime]:
        if "createdDate" in self and self["createdDate"] is not None:
            return read_nodejs_date(str(self["createdDate"]))
        else:
            return None

    @property
    def duration(self) -> int:
        return int(self["duration"])

    @property
    def system_name(self) -> str:
        return str(self["systemName"])

    @property
    def metadata_source(self) -> str:
        return str(self["metadataSource"])

    @property
    def properties(self) -> Optional[List[str]]:
        if "properties" in self and self["properties"] is not None:
            return list(self["properties"])
        else:
            return None

    @property
    def description(self) -> Optional[str]:
        if "description" in self and self["description"] is not None:
            return str(self["description"])
        else:
            return None

    @property
    def job(self) -> JobOutput:
        return self._job



from dataclasses import dataclass
from typing import Optional, List

from .artifact_type import ArtifactType
from .artifact_version import ArtifactVersion
from .code_version import GitVersionInput
from .files_metadata import FileMetadata
from .metric import MetricInput
from .model import ModelType
from .model_version import ModelVersionStatus
from .property import PropertyInput, PropertyOutput
from .user_declared_version import UserDeclaredVersion


@dataclass
class RulesDatasetVersionInput:
    parentName: Optional[str] = None
    parentId: Optional[int] = None
    autoVersion: Optional[bool] = None
    version: Optional[ArtifactVersion] = None
    userDeclaredVersion: Optional[UserDeclaredVersion] = None
    properties: Optional[List[PropertyInput]] = None
    dataResources: Optional[List[PropertyInput]] = None
    filesMetadata: Optional[List[FileMetadata]] = None
    description: Optional[str] = None


class ArtifactVersionOutput(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if "version" in self:
            self._version = ArtifactVersion(self["version"])
        if "properties" in self:
            self._properties = PropertyOutput.from_dict(self["properties"])


@dataclass
class RulesModelVersionInput:
    parentName: Optional[str] = None
    parentId: Optional[int] = None
    algorithmName: Optional[str] = None
    version: Optional[ArtifactVersion] = None
    userDeclaredVersion: Optional[UserDeclaredVersion] = None
    metrics: Optional[List[MetricInput]] = None
    properties: Optional[List[PropertyInput]] = None
    type: Optional[ModelType] = None
    status: Optional[ModelVersionStatus] = None
    description: Optional[str] = None


@dataclass
class RulesCodeVersionInput:
    parentName: Optional[str] = None
    parentId: Optional[int] = None
    version: Optional[ArtifactVersion] = None
    gitVersion: Optional[GitVersionInput] = None
    userDeclaredVersion: Optional[UserDeclaredVersion] = None


@dataclass
class ArtifactReferenceInput:
    artifactType: ArtifactType
    description: Optional[str] = None
    dataset: Optional[RulesDatasetVersionInput] = None
    model: Optional[RulesModelVersionInput] = None
    code: Optional[RulesCodeVersionInput] = None


class ArtifactReferenceOutput(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if "dataset" in self:
            self._dataset = RulesDatasetVersionInput(**self["dataset"])
        else:
            self._dataset = None
        if "model" in self:
            self._model = RulesModelVersionInput(**self["model"])
        else:
            self._model = None
        if "code" in self:
            self._code = RulesCodeVersionInput(**self["code"])
        else:
            self._code = None
